"""
String to type annotation map that find type annotation by method and argument name.
"""

from mypy_boto3_builder.constants import ALL
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_maps.named_unions import (
    ConditionBaseImportTypeDef,
    CopySourceOrStrTypeDef,
)
from mypy_boto3_builder.type_maps.typed_dicts import CopySourceTypeDef

__all__ = ("get_method_type_stub",)


ArgumentTypeMap = dict[str, FakeAnnotation]
MethodTypeMap = dict[str, ArgumentTypeMap]
ClassTypeMap = dict[str, MethodTypeMap]
ServiceTypeMap = dict[ServiceName, ClassTypeMap]


QueueAttributeFilterType = TypeLiteral(
    "QueueAttributeFilterType",
    [
        "All",
        "AWSTraceHeader",
        "ApproximateFirstReceiveTimestamp",
        "ApproximateReceiveCount",
        "MessageDeduplicationId",
        "MessageGroupId",
        "SenderId",
        "SentTimestamp",
        "SequenceNumber",
        "Policy",
        "VisibilityTimeout",
        "MaximumMessageSize",
        "MessageRetentionPeriod",
        "ApproximateNumberOfMessages",
        "ApproximateNumberOfMessagesNotVisible",
        "CreatedTimestamp",
        "LastModifiedTimestamp",
        "QueueArn",
        "ApproximateNumberOfMessagesDelayed",
        "DelaySeconds",
        "ReceiveMessageWaitTimeSeconds",
        "RedrivePolicy",
        "FifoQueue",
        "ContentBasedDeduplication",
        "KmsMasterKeyId",
        "KmsDataKeyReusePeriodSeconds",
        "DeduplicationScope",
        "FifoThroughputLimit",
        "RedriveAllowPolicy",
        "SqsManagedSseEnabled",
    ],
)

DEFAULT_VALUE_MAP: ServiceTypeMap = {
    ServiceNameCatalog.glacier: {
        "Client": {
            ALL: {
                "accountId": TypeConstant("-"),
            }
        }
    },
}

TYPE_MAP: ServiceTypeMap = {
    ServiceNameCatalog.s3: {
        # FIXME: boto3 overrides CopySource parameters for some S3 methods.
        # Types are set according to docs, might be incorrect
        "Client": {
            "copy_object": {"CopySource": CopySourceOrStrTypeDef},
            "upload_part_copy": {"CopySource": CopySourceOrStrTypeDef},
            "copy": {"CopySource": CopySourceTypeDef},
        },
        "MultipartUploadPart": {"copy_from": {"CopySource": CopySourceOrStrTypeDef}},
        "Bucket": {"copy": {"CopySource": CopySourceTypeDef}},
        "Object": {"copy": {"CopySource": CopySourceTypeDef}},
    },
    ServiceNameCatalog.dynamodb: {
        "Table": {
            "batch_writer": {
                "return": ExternalImport(ImportString("boto3", "dynamodb", "table"), "BatchWriter")
            },
            "query": {
                "KeyConditionExpression": ConditionBaseImportTypeDef,
                "FilterExpression": ConditionBaseImportTypeDef,
                "ConditionExpression": ConditionBaseImportTypeDef,
            },
            "delete_item": {
                "ConditionExpression": ConditionBaseImportTypeDef,
            },
            "put_item": {
                "ConditionExpression": ConditionBaseImportTypeDef,
            },
            "update_item": {
                "ConditionExpression": ConditionBaseImportTypeDef,
            },
            "scan": {
                "FilterExpression": ConditionBaseImportTypeDef,
            },
        },
    },
    ServiceNameCatalog.sqs: {
        ALL: {
            # FIXME: https://github.com/boto/botocore/issues/2726
            "receive_messages": {
                "AttributeNames": TypeSubscript(Type.Sequence, [QueueAttributeFilterType]),
            },
            "receive_message": {
                "AttributeNames": TypeSubscript(Type.Sequence, [QueueAttributeFilterType]),
            },
            "get_queue_attributes": {
                "AttributeNames": TypeSubscript(Type.Sequence, [QueueAttributeFilterType]),
            },
        }
    },
}


def _get_from_method_map(
    method_name: str,
    argument_name: str,
    method_type_map: MethodTypeMap,
) -> FakeAnnotation | None:
    if ALL in method_type_map:
        operation_type_map = method_type_map[ALL]
        if argument_name in operation_type_map:
            return operation_type_map[argument_name]

    if method_name in method_type_map:
        operation_type_map = method_type_map[method_name]
        if argument_name in operation_type_map:
            return operation_type_map[argument_name]

    return None


def _get_from_class_map(
    class_name: str,
    method_name: str,
    argument_name: str,
    class_type_map: ClassTypeMap,
) -> FakeAnnotation | None:
    if class_name in class_type_map:
        result = _get_from_method_map(method_name, argument_name, class_type_map[class_name])
        if result:
            return result
    if ALL in class_type_map:
        result = _get_from_method_map(method_name, argument_name, class_type_map[ALL])
        if result:
            return result
    return None


def _get_from_service_map(
    service_name: ServiceName,
    class_name: str,
    method_name: str,
    argument_name: str,
    service_type_map: ServiceTypeMap,
) -> FakeAnnotation | None:
    if service_name not in service_type_map:
        return None

    return _get_from_class_map(
        class_name,
        method_name,
        argument_name,
        service_type_map[service_name],
    )


def get_method_type_stub(
    service_name: ServiceName, class_name: str, method_name: str, argument_name: str
) -> FakeAnnotation | None:
    """
    Get stub type for method argument.

    Arguments:
        service_name -- Service name.
        class_name -- Parent class name.
        method_name -- Method name.
        argument_name -- Argument name.

    Returns:
        Type annotation or None.
    """
    return _get_from_service_map(service_name, class_name, method_name, argument_name, TYPE_MAP)


def get_default_value_stub(
    service_name: ServiceName, class_name: str, method_name: str, argument_name: str
) -> TypeConstant | None:
    """
    Get default value stub for method argument.

    Arguments:
        service_name -- Service name.
        class_name -- Parent class name.
        method_name -- Method name.
        argument_name -- Argument name.

    Returns:
        TypeConstant or None.
    """
    result = _get_from_service_map(
        service_name, class_name, method_name, argument_name, DEFAULT_VALUE_MAP
    )
    if result is None:
        return None
    if not isinstance(result, TypeConstant):
        return None
    return result



"""
Postprocessor for all classes and methods.
"""
from mypy_boto3_builder.postprocessors.base import BasePostprocessor


class BotocorePostprocessor(BasePostprocessor):
    """
    Postprocessor for botocore classes and methods.
    """

    def process_package(self) -> None:
        """
        Leave package as it is.
        """



from types import TracebackType
from typing import Any, List, Optional, Type, Union

from aiobotocore.client import AioBaseClient as AioBaseClient
from aiobotocore.client import AioClientCreator as AioClientCreator
from aiobotocore.credentials import AioCredentials as AioCredentials
from aiobotocore.credentials import create_credential_resolver as create_credential_resolver
from aiobotocore.hooks import AioHierarchicalEmitter as AioHierarchicalEmitter
from aiobotocore.parsers import AioResponseParserFactory as AioResponseParserFactory
from botocore.config import Config
from botocore.model import ServiceModel
from botocore.session import EVENT_ALIASES as EVENT_ALIASES
from botocore.session import Session

class ClientCreatorContext:
    def __init__(self, coro: Any) -> None: ...
    async def __aenter__(self) -> AioBaseClient: ...
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None: ...

class AioSession(Session):
    def __init__(
        self,
        session_vars: Optional[Any] = ...,
        event_hooks: Optional[Any] = ...,
        include_builtin_handlers: bool = ...,
        profile: Optional[Any] = ...,
    ) -> None: ...
    def register(
        self,
        event_name: str,
        handler: Any,
        unique_id: Optional[Any] = ...,
        unique_id_uses_count: bool = ...,
    ) -> None: ...
    def create_client(  # type: ignore [override]
        self,
        service_name: str,
        region_name: Optional[str] = ...,
        api_version: Optional[str] = ...,
        use_ssl: Optional[bool] = ...,
        verify: Union[bool, str, None] = ...,
        endpoint_url: Optional[str] = ...,
        aws_access_key_id: Optional[str] = ...,
        aws_secret_access_key: Optional[str] = ...,
        aws_session_token: Optional[str] = ...,
        config: Optional[Config] = ...,
    ) -> AioBaseClient: ...
    async def get_credentials(self) -> AioCredentials: ...  # type: ignore [override]
    def set_credentials(
        self, access_key: str, secret_key: str, token: Optional[Any] = ...
    ) -> None: ...
    async def get_service_model(  # type: ignore [override]
        self, service_name: str, api_version: Optional[Any] = ...
    ) -> ServiceModel: ...
    async def get_service_data(
        self, service_name: str, api_version: Optional[Any] = ...
    ) -> Any: ...
    async def get_available_regions(  # type: ignore [override]
        self, service_name: str, partition_name: str = ..., allow_non_regional: bool = ...
    ) -> List[str]: ...

def get_session(env_vars: Optional[Any] = ...) -> AioSession: ...



import logging
from types import TracebackType
from typing import Optional, Type, TypeVar

from boto3.resources.base import ServiceResource

_R = TypeVar("_R", bound="AIOBoto3ServiceResource")

logger: logging.Logger

class AIOBoto3ServiceResource(ServiceResource):
    async def __aenter__(self: _R) -> _R: ...
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None: ...
    def close(self) -> None: ...



"""
Annotation to mark argument for removal.
"""
from typing import TypeVar

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation

_R = TypeVar("_R", bound="RemoveArgument")


class RemoveArgument(FakeAnnotation):
    """
    Annotation to mark argument for removal.
    """

    def render(self, parent_name: str = "") -> str:
        """
        Not used.
        """
        return "None"

    def copy(self: _R) -> _R:
        """
        Not used.
        """
        return self



from types import TracebackType
from typing import Any, Optional, Type, TypeVar, Union

from aiobotocore.paginate import AioPaginator
from aiobotocore.waiter import AIOWaiter
from botocore.client import BaseClient, ClientCreator
from botocore.config import Config
from botocore.history import HistoryRecorder

history_recorder: HistoryRecorder

_R = TypeVar("_R")

class AioClientCreator(ClientCreator):
    async def create_client(  # type: ignore [override]
        self,
        service_name: str,
        region_name: str,
        is_secure: bool = ...,
        endpoint_url: Optional[str] = ...,
        verify: Optional[Union[str, bool]] = ...,
        credentials: Optional[Any] = ...,
        scoped_config: Optional[Any] = ...,
        api_version: Optional[str] = ...,
        client_config: Optional[Config] = ...,
        auth_token: Optional[str] = ...,
    ) -> AioBaseClient: ...

class AioBaseClient(BaseClient):
    def get_paginator(self, operation_name: str) -> AioPaginator: ...
    def get_waiter(self, waiter_name: str) -> AIOWaiter: ...
    async def __aenter__(self: _R) -> _R: ...
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> Any: ...
    async def close(self) -> None: ...  # type: ignore [override]



from typing import Any, Optional

from botocore.config import Config
from botocore.retryhandler import CRC32Checker, MaxAttemptsDecorator, MultiChecker, RetryHandler

def create_retry_handler(
    config: Config, operation_name: Optional[str] = ...
) -> AioRetryHandler: ...
def create_checker_from_retry_config(
    config: Config, operation_name: Optional[str] = ...
) -> AioMaxAttemptsDecorator: ...

class AioRetryHandler(RetryHandler):
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

class AioMaxAttemptsDecorator(MaxAttemptsDecorator):
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

class AioMultiChecker(MultiChecker):
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

class AioCRC32Checker(CRC32Checker):
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...



import datetime
from logging import Logger
from typing import Any, Callable

from botocore.session import Session
from botocore.tokens import (
    DeferredRefreshableToken,
    FrozenAuthToken,
    SSOTokenProvider,
    TokenProviderChain,
)

logger: Logger

def create_token_resolver(session: Session) -> TokenProviderChain: ...

class AioDeferredRefreshableToken(DeferredRefreshableToken):
    def __init__(
        self,
        method: Any,
        refresh_using: Callable[[], FrozenAuthToken],
        time_fetcher: Callable[[], datetime.datetime] = ...,
    ) -> None: ...
    async def get_frozen_token(self) -> FrozenAuthToken: ...  # type: ignore [override]

class AioSSOTokenProvider(SSOTokenProvider):
    def load_token(self) -> AioDeferredRefreshableToken: ...



"""
Sorter for TypeDefSorter to prevent import errors.
"""
from collections.abc import Iterable
from graphlib import CycleError, TopologicalSorter

from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.type_annotations.type_def_sortable import TypeDefSortable


class TypeDefSorter:
    """
    Sorter for TypeDefs to prevent import errors.
    """

    ATTEMPTS = 20

    def __init__(self, type_defs: Iterable[TypeDefSortable]) -> None:
        self.typed_defs = list(type_defs)
        self.typed_defs.sort(key=lambda x: x.name)
        self.typed_def_map = self._get_type_defs_map(self.typed_defs)
        self.logger = get_logger()

    @classmethod
    def _get_type_defs_map(
        cls, type_defs: Iterable[TypeDefSortable], skip: Iterable[str] = ()
    ) -> dict[str, TypeDefSortable]:
        result: dict[str, TypeDefSortable] = {}
        for type_def in type_defs:
            result[type_def.name] = type_def
            children = type_def.get_sortable_children()
            child_skip = {*skip, *result.keys()}
            new_children = [i for i in children if i.name not in child_skip]
            if new_children:
                result.update(cls._get_type_defs_map(new_children, skip=child_skip))

        return result

    def sort(self) -> list[TypeDefSortable]:
        """
        Sort items with TopologicalSorter or stringify as a fallback.
        """
        for attempt in range(self.ATTEMPTS):
            try:
                result = self._sort_topological()
                if attempt:
                    self.logger.debug(f"Topological sort succeeded after {attempt + 1} attempts")
                return result
            except CycleError as e:
                for type_def in self._get(*e.args[-1]):
                    self.logger.debug(
                        f"Stringifying {type_def.name}: unsortable children"
                        f" {self.typed_def_map[type_def.name]}"
                    )
                    type_def.stringify()

        self.logger.warn("Topological sort failed, stringifying TypedDicts")
        return self._sort_stringified()

    def _get(self, *names: str) -> list[TypeDefSortable]:
        result: list[TypeDefSortable] = []
        for name in names:
            type_def = self.typed_def_map[name]
            if type_def in result:
                continue

            result.append(type_def)

        return result

    @staticmethod
    def _get_children_names(type_def: TypeDefSortable) -> set[str]:
        result: set[str] = set()
        for child in type_def.get_sortable_children():
            if child.is_stringified():
                continue
            result.add(child.name)

        return result

    def _create_graph(self) -> dict[str, list[str]]:
        result: dict[str, list[str]] = {}
        for name in sorted(self.typed_def_map):
            result[name] = sorted(self._get_children_names(self.typed_def_map[name]))

        return result

    def _sort_topological(self) -> list[TypeDefSortable]:
        graph = self._create_graph()
        names_sorted = list(TopologicalSorter(graph).static_order())
        result = self._get(*names_sorted)
        return result

    def _sort_stringified(self) -> list[TypeDefSortable]:
        added_hashes: list[int] = []
        result: list[TypeDefSortable] = []
        discovered: list[TypeDefSortable] = []

        for type_annotation in sorted(self.typed_defs):
            if hash(type_annotation) in added_hashes:
                continue

            result.append(type_annotation)
            added_hashes.append(hash(type_annotation))
            discovered.append(type_annotation)

        while discovered:
            child = discovered.pop()
            sub_typed_defs = child.get_sortable_children()
            for child_type_def in sorted(sub_typed_defs):
                child_type_def.stringify()

                if hash(child_type_def) in added_hashes:
                    continue

                result.append(child_type_def)
                added_hashes.append(hash(child_type_def))
                discovered.append(child_type_def)

        result.sort()
        return result



"""
Version-related utils.
"""
import contextlib
import importlib.metadata

from boto3 import __version__ as boto3_version
from botocore import __version__ as botocore_version
from newversion import Version

from mypy_boto3_builder.constants import PACKAGE_NAME


def get_builder_version() -> str:
    """
    Get program version.
    """
    with contextlib.suppress(importlib.metadata.PackageNotFoundError):
        return importlib.metadata.version(PACKAGE_NAME)

    return "0.0.0"


def get_min_build_version(version: str) -> str:
    """
    Get min version build version by setting micro to 0.
    """
    return Version(version).replace(micro=0).get_stable().dumps()


def get_max_build_version(version: str) -> str:
    """
    Get min version build version by bumping minor.
    """
    return Version(version).bump_minor().get_stable().dumps()


def get_botocore_version() -> str:
    """
    Get botocore package version.
    """
    return botocore_version


def get_boto3_version() -> str:
    """
    Get boto3 package version.
    """
    return boto3_version


def get_aiobotocore_version() -> str:
    """
    Get aiobotocore package version.
    """
    try:
        from aiobotocore import __version__ as version  # type: ignore
    except ImportError:
        raise RuntimeError("aiobotocore is not installed")
    return f"{version}"


def get_aioboto3_version() -> str:
    """
    Get aioboto3 package version.
    """
    try:
        from aioboto3 import __version__ as version  # type: ignore
    except ImportError:
        raise RuntimeError("aioboto3 is not installed")

    return f"{version}"



"""
Processors for parsing and writing `aioboto3` modules.
"""
from collections.abc import Iterable
from pathlib import Path

from boto3.session import Session

from mypy_boto3_builder.constants import AIOBOTO3_STUBS_STATIC_PATH, TEMPLATES_PATH
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.package_data import TypesAioBoto3LitePackageData, TypesAioBoto3PackageData
from mypy_boto3_builder.parsers.types_aioboto3_package import parse_types_aioboto3_package
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.types_aioboto3_package import TypesAioBoto3Package
from mypy_boto3_builder.utils.nice_path import NicePath
from mypy_boto3_builder.writers.package_writer import PackageWriter


def process_types_aioboto3(
    session: Session,
    output_path: Path,
    service_names: Iterable[ServiceName],
    generate_setup: bool,
    version: str,
) -> TypesAioBoto3Package:
    """
    Parse and write stubs package `types-aioboto3`.

    Arguments:
        session -- boto3 session
        output_path -- Package output path
        service_names -- List of known service names
        generate_setup -- Generate ready-to-install or to-use package
        version -- Package version

    Return:
        Parsed TypesAioBoto3Package.
    """
    logger = get_logger()
    package = parse_types_aioboto3_package(session, service_names, TypesAioBoto3PackageData)
    package.version = version
    logger.debug(f"Writing {package.pypi_name} to {NicePath(output_path)}")

    package_writer = PackageWriter(output_path=output_path, generate_setup=generate_setup)
    package_writer.write_package(
        package,
        templates_path=TEMPLATES_PATH / "types-aioboto3",
        static_files_path=AIOBOTO3_STUBS_STATIC_PATH,
    )
    return package


def process_types_aioboto3_lite(
    session: Session,
    output_path: Path,
    service_names: Iterable[ServiceName],
    generate_setup: bool,
    version: str,
) -> TypesAioBoto3Package:
    """
    Parse and write stubs package `types-aioboto3-lite`.

    Arguments:
        session -- boto3 session
        output_path -- Package output path
        service_names -- List of known service names
        generate_setup -- Generate ready-to-install or to-use package
        version -- Package version

    Return:
        Parsed AioBotocoreStubsPackage.
    """
    logger = get_logger()
    package = parse_types_aioboto3_package(session, service_names, TypesAioBoto3LitePackageData)
    package.version = version
    logger.debug(f"Writing {package.pypi_name} to {NicePath(output_path)}")

    package_writer = PackageWriter(output_path=output_path, generate_setup=generate_setup)
    package_writer.write_package(
        package,
        templates_path=TEMPLATES_PATH / "types-aioboto3",
        static_files_path=AIOBOTO3_STUBS_STATIC_PATH,
        exclude_template_names=["session.pyi.jinja2"],
    )
    return package


def process_types_aioboto3_docs(
    session: Session,
    output_path: Path,
    service_names: Iterable[ServiceName],
) -> TypesAioBoto3Package:
    """
    Parse and write master package docs.

    Arguments:
        session -- boto3 session
        output_path -- Package output path
        service_names -- List of known service names

    Return:
        Parsed AioBotocoreStubsPackage.
    """
    logger = get_logger()
    package = parse_types_aioboto3_package(session, service_names, TypesAioBoto3PackageData)

    logger.debug(f"Writing {package.pypi_name} to {NicePath(output_path)}")

    package_writer = PackageWriter(output_path=output_path)
    package_writer.write_docs(
        package,
        templates_path=TEMPLATES_PATH / "types_aioboto3_docs",
    )

    return package



import logging
from typing import Any

from aioboto3.session import Session as Session

__author__: str
__email__: str
__version__: str

class NullHandler(logging.Handler):
    def emit(self, record: Any) -> Any: ...



import logging
from typing import Any, AsyncIterator, List, NoReturn

from boto3.resources.base import ServiceResource
from boto3.resources.collection import CollectionFactory, CollectionManager, ResourceCollection
from boto3.resources.factory import ResourceFactory
from boto3.resources.model import Collection
from boto3.utils import ServiceContext
from botocore.hooks import BaseEventHooks

logger: logging.Logger

class AIOResourceCollection(ResourceCollection):
    """
    Converted the ResourceCollection.pages() function to an async generator so that we can do
    async for on a paginator inside that function

    Converted the __iter__
    """

    def __anext__(self) -> AsyncIterator[Any]: ...
    def __aiter__(self) -> AsyncIterator[Any]: ...
    def __iter__(self) -> NoReturn: ...
    def pages(self) -> AsyncIterator[List[Any]]: ...  # type: ignore [override]

class AIOCollectionManager(CollectionManager):
    def __init__(
        self,
        collection_model: Collection,
        parent: ServiceResource,
        factory: ResourceFactory,
        service_context: ServiceContext,
    ) -> None: ...

class AIOCollectionFactory(CollectionFactory):
    def load_from_definition(
        self,
        resource_name: str,
        collection_model: Collection,
        service_context: ServiceContext,
        event_emitter: BaseEventHooks,
    ) -> AIOCollectionManager: ...



from typing import Any, Dict, Mapping, TypeVar

from botocore.awsrequest import AWSHTTPResponse, AWSRequest
from botocore.httpchecksum import AwsChunkedWrapper
from botocore.model import OperationModel

_R = TypeVar("_R")

class AioAwsChunkedWrapper(AwsChunkedWrapper):
    async def _make_chunk(self) -> bytes: ...
    def __aiter__(self: _R) -> _R: ...
    async def __anext__(self) -> bytes: ...

async def handle_checksum_body(
    http_response: AWSHTTPResponse,
    response: Dict[str, Any],
    context: Mapping[str, Any],
    operation_model: OperationModel,
) -> None: ...
def apply_request_checksum(request: AWSRequest) -> None: ...



"""
Wrapper for `typing/typing_extensions.TypedDict` type annotations.
"""
from collections.abc import Iterable, Iterator
from typing import TypeVar

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.internal_import_record import InternalImportRecord
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_def_sortable import TypeDefSortable
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript

_R = TypeVar("_R", bound="TypeTypedDict")


class TypedDictAttribute:
    """
    TypedDict attribute wrapper.

    Arguments:
        name -- Attribute name.
        type_annotation -- Attribute type annotation.
        required -- Whether the attribute has to be set.
    """

    def __init__(self, name: str, type_annotation: FakeAnnotation, required: bool) -> None:
        self.name = name
        self.required = required
        self.type_annotation = type_annotation

    def __hash__(self) -> int:
        return hash((self.name, self.required, self.type_annotation.get_sort_key()))

    def get_type_annotation(self) -> FakeAnnotation:
        """
        Get wrapped for non-required type annotation or raw type annotation.
        """
        if self.is_required():
            return self.type_annotation

        return TypeSubscript(Type.NotRequired, [self.type_annotation])

    def render(self, parent_name: str = "") -> str:
        """
        Render attribute to use in class-based TypedDict definition.

        Returns:
            A string with argument definition.
        """
        return f'"{self.name}": {self.get_type_annotation().render(parent_name)}'

    def iterate_types(self) -> Iterator[FakeAnnotation]:
        """
        Itera over type annotations.
        """
        yield from self.type_annotation.iterate_types()

    def is_required(self) -> bool:
        """
        Whether argument is required.
        """
        return self.required


class TypeTypedDict(FakeAnnotation, TypeDefSortable):
    """
    Wrapper for `typing/typing_extensions.TypedDict` type annotations.

    Arguments:
        name -- Type name.
        children -- Typed dict attributes.
        docstring -- Docstring for render.
        stringify -- Convert type annotation to string to avoid circular deps.
    """

    def __init__(
        self,
        name: str,
        children: Iterable[TypedDictAttribute] = (),
        docstring: str = "",
        stringify: bool = False,
    ) -> None:
        self.name = name
        self.children = list(children)
        self.docstring = docstring
        self._stringify = stringify

    def is_stringified(self) -> bool:
        """
        Whether TypedDict usage should be rendered as a string.
        """
        return self._stringify

    def stringify(self) -> None:
        """
        Render TypedDict usage as a string.
        """
        self._stringify = True

    def get_sort_key(self) -> str:
        """
        Sort Typed Dicts by name.
        """
        return self.name

    def __hash__(self) -> int:
        return hash((self.name, *[hash(i) for i in self.children]))

    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        if self.is_stringified():
            return f'"{self.name}"'

        if parent_name and parent_name == self.name:
            return f'"{self.name}"'

        return self.name

    def debug_render(self) -> str:
        """
        Render type annotation for debug purposes.
        """
        return f"{self}: {', '.join([c.render() for c in self.children])}"

    def get_definition_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using TypedDict.
        """
        result = Type.TypedDict.get_import_records()
        for child in self.iterate_children():
            result.update(child.get_type_annotation().get_import_records())
        return result

    def _get_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using type annotation.
        """
        return {InternalImportRecord(ServiceModuleName.type_defs, name=self.name)}

    def add_attribute(self, name: str, type_annotation: FakeAnnotation, required: bool) -> None:
        """
        Add new attribute to a dictionary.

        Arguments:
            name -- Argument name.
            type_annotation -- Argument type annotation.
            required -- Whether argument has to be set.
        """
        self.children.append(TypedDictAttribute(name, type_annotation, required))

    def is_dict(self) -> bool:
        """
        Whether type annotation is `Dict` or `TypedDict`.
        """
        return True

    def is_typed_dict(self) -> bool:
        """
        Whether type annotation is `TypedDict`.
        """
        return True

    def has_optional(self) -> bool:
        """
        Whether TypedDict has optional keys.
        """
        return any(not child.is_required() for child in self.children)

    def has_required(self) -> bool:
        """
        Whether TypedDict has required keys.
        """
        return any(child.is_required() for child in self.children)

    def has_both(self) -> bool:
        """
        Whether TypedDict has both optional and required keys.
        """
        return self.has_required() and self.has_optional()

    def get_required(self) -> list[TypedDictAttribute]:
        """
        Get a list of required attributes.
        """
        result: list[TypedDictAttribute] = []
        for child in self.children:
            if child.is_required():
                result.append(child)
        return result

    def get_optional(self) -> list[TypedDictAttribute]:
        """
        Get a list of optional attributes.
        """
        result: list[TypedDictAttribute] = []
        for child in self.children:
            if not child.is_required():
                result.append(child)
        return result

    def copy(self: _R) -> _R:
        """
        Create a copy of type annotation wrapper.
        """
        return self.__class__(
            self.name,
            list(self.children),
            docstring=self.docstring,
            stringify=self.is_stringified(),
        )

    def is_same(self: _R, other: _R) -> bool:
        """
        Check whether typed dict attributes are the same as `other`.
        """
        return hash(self) == hash(other)

    def get_children_types(self) -> set[FakeAnnotation]:
        """
        Extract required type annotations from attributes.
        """
        result: set[FakeAnnotation] = set()
        for child in self.children:
            result.update(child.iterate_types())
        return result

    def get_sortable_children(self) -> list[TypeDefSortable]:
        """
        Extract required TypeDefSortable list from attributes.
        """
        result: list[TypeDefSortable] = []
        children_types = self.get_children_types()
        for type_annotation in children_types:
            if not isinstance(type_annotation, TypeDefSortable):
                continue
            result.append(type_annotation)

        return result

    def get_children_literals(self, processed: Iterable[str] = ()) -> set[TypeLiteral]:
        """
        Extract required TypeLiteral list from attributes.
        """
        result: set[TypeLiteral] = set()
        if self.name in processed:
            return result
        children_types = self.get_children_types()
        for type_annotation in children_types:
            if isinstance(type_annotation, TypeLiteral):
                result.add(type_annotation)
            if isinstance(type_annotation, TypeDefSortable):
                result.update(type_annotation.get_children_literals((self.name, *processed)))
        return result

    def iterate_children(self) -> Iterator[TypedDictAttribute]:
        """
        Iterate over children from required to optional.
        """
        yield from self.get_required()
        yield from self.get_optional()

    def get_local_types(self) -> list[FakeAnnotation]:
        """
        Get internal types generated by builder.
        """
        return [self]

    @property
    def type_hint_annotations(self) -> list[FakeAnnotation]:
        """
        Type annotations list from arguments and return type with internal types.
        """
        result: list[FakeAnnotation] = []
        for child in self.children:
            if child.type_annotation.get_local_types():
                result.append(child.type_annotation)
        return result

    def is_type_def(self) -> bool:
        """
        Whether type annotation is a TypeDef.
        """
        return True

    def is_union(self) -> bool:
        """
        Whether type annotation is a TypeUnion.
        """
        return False



"""
Parser for boto3 changelog.
"""
import re
from urllib.error import HTTPError
from urllib.request import urlopen


class BotocoreChangelog:
    """
    Parser for boto3 changelog.
    """

    URL = "https://raw.githubusercontent.com/boto/botocore/develop/CHANGELOG.rst"
    SERVICE_NAME_RE = re.compile(r"``(\S+)``")

    def __init__(self) -> None:
        self._changelog: str | None = None

    def _get_changelog(self) -> str:
        if self._changelog is not None:
            return self._changelog

        try:
            with urlopen(self.URL) as response:
                self._changelog = response.read().decode()
        except HTTPError as e:
            raise RuntimeError(f"Cannot retrieve {self.URL}: {e}") from None

        return self._changelog

    def _get_section(self, version: str) -> str:
        result: list[str] = []
        changelog = self._get_changelog()
        found = False
        for line in changelog.splitlines():
            if line == version:
                found = True
                continue
            if found and result and line.startswith("==="):
                result.pop()
                break
            if found:
                result.append(line)

        return "\n".join(result[1:])

    def get_updated_service_names(self, version: str) -> list[str]:
        """
        Get a list of service names updated in `version` release.
        """
        result: list[str] = []
        section = self._get_section(version)
        for match in self.SERVICE_NAME_RE.findall(section):
            if match not in result:
                result.append(match)

        return result



"""
RDS service injected methods.
"""


from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript

generate_db_auth_token_method = Method(
    "generate_db_auth_token",
    [
        Argument("self", None),
        Argument("DBHostname", Type.str),
        Argument("Port", Type.int),
        Argument("DBUsername", Type.str),
        Argument("Region", TypeSubscript(Type.Optional, [Type.str]), Type.Ellipsis),
    ],
    Type.str,
)

CLIENT_METHODS = [
    generate_db_auth_token_method,
]



"""
String to type annotation map to replace overriden botocore shapes.
"""
from collections.abc import Iterable

from mypy_boto3_builder.constants import ALL
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_maps.named_unions import (
    AttributeValueTypeDef,
    BlobTypeDef,
    PolicyDocumentTypeDef,
    StreamingBodyType,
    TableAttributeValueTypeDef,
    TimestampTypeDef,
    UniversalAttributeValueTypeDef,
)
from mypy_boto3_builder.type_maps.typed_dicts import GetTemplateOutputTypeDef

ShapeTypeMap = dict[ServiceName, dict[str, dict[str, FakeAnnotation]]]

SHAPE_TYPE_MAP: ShapeTypeMap = {
    ServiceNameCatalog.all: {
        ALL: {
            "integer": Type.int,
            "long": Type.int,
            "boolean": Type.bool,
            "double": Type.float,
            "float": Type.float,
            "timestamp": TimestampTypeDef,
            "blob": BlobTypeDef,
            "blob_streaming": BlobTypeDef,
        }
    },
    ServiceNameCatalog.dynamodb: {
        ALL: {
            "AttributeValueTypeDef": UniversalAttributeValueTypeDef,
            "ConditionExpressionTypeDef": Type.bool,
        },
        "ServiceResource": {
            "AttributeValueTypeDef": TableAttributeValueTypeDef,
        },
        "Table": {
            "AttributeValueTypeDef": TableAttributeValueTypeDef,
        },
    },
}

OUTPUT_SHAPE_TYPE_MAP: ShapeTypeMap = {
    ServiceNameCatalog.all: {
        ALL: {
            "timestamp": Type.datetime,
            "blob": Type.bytes,
            "blob_streaming": StreamingBodyType,
        },
    },
    ServiceNameCatalog.iam: {
        ALL: {
            "policyDocumentType": PolicyDocumentTypeDef,
        },
    },
    ServiceNameCatalog.dynamodb: {
        ALL: {
            "AttributeValueTypeDef": AttributeValueTypeDef,
        },
        "ServiceResource": {
            "AttributeValueTypeDef": TableAttributeValueTypeDef,
        },
        "Table": {
            "AttributeValueTypeDef": TableAttributeValueTypeDef,
        },
    },
    # FIXME: botocore processes TemplateBody with json_decode_template_body
    ServiceNameCatalog.cloudformation: {
        ALL: {
            "GetTemplateOutputTypeDef": GetTemplateOutputTypeDef,
        },
    },
}


def get_shape_type_stub(
    shape_type_maps: Iterable[ShapeTypeMap],
    service_name: ServiceName,
    resource_name: str,
    shape_name: str,
) -> FakeAnnotation | None:
    """
    Get stub type for input botocore shape.

    Arguments:
        shape_type_map -- Map to lookup
        service_name -- Service name
        resource_name -- Resource name
        shape_name -- Target Shape name

    Returns:
        Type annotation or None.
    """
    checks = (
        (service_name, resource_name),
        (service_name, ALL),
        (ServiceNameCatalog.all, resource_name),
        (ServiceNameCatalog.all, ALL),
    )
    for shape_type_map in shape_type_maps:
        for current_service_name, current_resource_name in checks:
            if current_service_name not in shape_type_map:
                continue
            service_type_map = shape_type_map[current_service_name]
            if current_resource_name not in service_type_map:
                continue
            resource_type_map = service_type_map[current_resource_name]
            if shape_name not in resource_type_map:
                continue
            return resource_type_map[shape_name]

    return None



"""
Parent class for all type annotation wrappers.
"""
from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import TypeVar

from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.import_string import ImportString

_R = TypeVar("_R", bound="FakeAnnotation")


class FakeAnnotation(ABC):
    """
    Parent class for all type annotation wrappers.
    """

    _sys_import_record = ImportRecord(ImportString("sys"))

    def __hash__(self) -> int:
        return hash(self.render())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FakeAnnotation):
            return False

        return self.get_sort_key() == other.get_sort_key()

    def __lt__(self, other: "FakeAnnotation") -> bool:
        return self.get_sort_key() < other.get_sort_key()

    def __gt__(self, other: "FakeAnnotation") -> bool:
        return self.get_sort_key() > other.get_sort_key()

    def get_sort_key(self) -> str:
        """
        Get string to sort annotations.
        """
        return str(self)

    def __str__(self) -> str:
        return self.render()

    @abstractmethod
    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.
        """

    def _get_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using type annotation.
        """
        return set()

    def get_import_records(self) -> set[ImportRecord]:
        """
        Get all import records required for using type annotation.
        """
        result: set[ImportRecord] = set()
        import_records = self._get_import_records()
        for import_record in import_records:
            if not import_record.is_empty() and not import_record.is_builtins():
                result.add(import_record)
            if import_record.needs_sys_fallback():
                result.add(self._sys_import_record)

        return result

    def iterate_types(self) -> Iterator["FakeAnnotation"]:
        """
        Iterate over all used type annotations recursively including self.
        """
        yield self

    def is_dict(self) -> bool:
        """
        Whether type annotation is `Dict` or `TypedDict`.
        """
        return False

    def is_typed_dict(self) -> bool:
        """
        Whether type annotation is `TypedDict`.
        """
        return False

    def is_list(self) -> bool:
        """
        Whether type annotation is `List`.
        """
        return False

    def is_literal(self) -> bool:
        """
        Whether type annotation is `Literal`.
        """
        return False

    @abstractmethod
    def copy(self: _R) -> _R:
        """
        Create a copy of type annotation wrapper.
        """

    def get_local_types(self) -> list["FakeAnnotation"]:
        """
        Get internal types generated by builder.
        """
        return []



"""
Postprocessor for aiobotocore classes and methods.
"""

from collections.abc import Iterator

from boto3.dynamodb.table import BatchWriter, TableResource
from boto3.resources.base import ServiceResource
from boto3.resources.collection import ResourceCollection
from botocore.client import BaseClient
from botocore.config import Config
from botocore.eventstream import EventStream
from botocore.paginate import Paginator
from botocore.response import StreamingBody
from botocore.waiter import Waiter

from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.postprocessors.base import BasePostprocessor
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.collection import Collection
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_def_sortable import TypeDefSortable
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_maps.aio_resource_method_map import get_aio_resource_method


class AioBotocorePostprocessor(BasePostprocessor):
    """
    Postprocessor for aiobotocore classes and methods.
    """

    COMMON_COLLECTION_METHOD_NAMES = (
        "__iter__",
        "__aiter__",
        "all",
        "pages",
        "filter",
        "limit",
        "page_size",
    )

    EXTERNAL_IMPORTS_MAP = {
        ExternalImport.from_class(StreamingBody): ExternalImport(
            ImportString("aiobotocore", "response"), "StreamingBody"
        ),
        ExternalImport.from_class(EventStream): ExternalImport(
            ImportString("aiobotocore", "eventstream"), "AioEventStream"
        ),
        ExternalImport.from_class(Config): ExternalImport(
            ImportString("aiobotocore", "config"), "AioConfig"
        ),
        ExternalImport.from_class(Waiter): ExternalImport(
            ImportString("aiobotocore", "waiter"), "AIOWaiter"
        ),
        ExternalImport.from_class(Paginator): ExternalImport(
            ImportString("aiobotocore", "paginate"), "AioPaginator"
        ),
        ExternalImport.from_class(BaseClient): ExternalImport(
            ImportString("aiobotocore", "client"), "AioBaseClient"
        ),
        ExternalImport.from_class(ServiceResource): ExternalImport(
            ImportString("aioboto3", "resources", "base"), "AIOBoto3ServiceResource", safe=True
        ),
        ExternalImport.from_class(ResourceCollection): ExternalImport(
            ImportString("aioboto3", "resources", "collection"), "AIOResourceCollection", safe=True
        ),
        ExternalImport.from_class(BatchWriter): ExternalImport(
            ImportString("aioboto3", "dynamodb", "table"), "BatchWriter", safe=True
        ),
        ExternalImport.from_class(TableResource): ExternalImport(
            ImportString("aioboto3", "dynamodb", "table"), "CustomTableResource", safe=True
        ),
    }

    def process_package(self) -> None:
        """
        Convert all methods to asynchronous.
        """
        self._make_async_client()
        self._make_async_paginators()
        self._make_async_waiters()
        self._make_async_service_resource()
        self._make_async_collections()
        self._make_async_sub_resources()
        self._add_contextmanager_methods()
        self._replace_external_imports()

    def _make_async_client(self) -> None:
        for method in self.package.client.methods:
            if method.name in [
                "exceptions",
                "get_waiter",
                "get_paginator",
                "can_paginate",
            ]:
                continue
            method.is_async = True

    @classmethod
    def _make_async_collection(cls, collection: Collection) -> None:
        for method in collection.methods:
            if method.name not in cls.COMMON_COLLECTION_METHOD_NAMES:
                method.is_async = True

        pages_method = collection.get_method("pages")
        assert isinstance(pages_method.return_type, TypeSubscript)
        pages_method.return_type.parent = Type.AsyncIterator

        aiter_method = collection.get_method("__iter__").copy()
        aiter_method.name = "__aiter__"
        assert isinstance(aiter_method.return_type, TypeSubscript)
        aiter_method.return_type.parent = Type.AsyncIterator
        collection.methods.append(aiter_method)

        iter_method = collection.get_method("__iter__")
        iter_method.return_type = Type.NoReturn

    def _make_async_collections(self) -> None:
        if not self.package.service_resource:
            return
        for collection in self.package.service_resource.collections:
            self._make_async_collection(collection)

    def _make_async_paginators(self) -> None:
        for paginator in self.package.paginators:
            paginate_method = paginator.get_method("paginate")
            assert isinstance(paginate_method.return_type, TypeSubscript)
            paginate_method.return_type.parent = Type.AsyncIterator

    def _make_async_waiters(self) -> None:
        for waiter in self.package.waiters:
            for method in waiter.methods:
                method.is_async = True

    def _make_async_service_resource(self) -> None:
        if not self.package.service_resource:
            return
        for method in self.package.service_resource.methods:
            method.is_async = True

    def _make_async_sub_resources(self) -> None:
        if not self.package.service_resource:
            return
        for sub_resource in self.package.service_resource.sub_resources:
            for method in sub_resource.methods:
                method_stub = get_aio_resource_method(
                    self.package.service_name, sub_resource.name, method.name
                )
                if method_stub:
                    method.arguments = method_stub.arguments
                    method.return_type = method_stub.return_type
                    method.is_async = method_stub.is_async
                    continue
                method.is_async = True
            for collection in sub_resource.collections:
                self._make_async_collection(collection)
            for attribute in sub_resource.attributes:
                if not attribute.is_autoload_property():
                    continue
                attribute.type_annotation = TypeSubscript(
                    Type.Awaitable, [attribute.type_annotation]
                )

    def _add_contextmanager_methods(self) -> None:
        self.package.client.methods.append(
            Method(
                "__aenter__",
                [
                    Argument("self", None),
                ],
                return_type=InternalImport(self.package.client.name),
                is_async=True,
                docstring=self.package.client.docstring,
            )
        )
        self.package.client.methods.append(
            Method(
                "__aexit__",
                [
                    Argument("self", None),
                    Argument("exc_type", Type.Any),
                    Argument("exc_val", Type.Any),
                    Argument("exc_tb", Type.Any),
                ],
                return_type=Type.Any,
                is_async=True,
                docstring=self.package.client.docstring,
            )
        )

    def _iterate_types_shallow(self) -> Iterator[FakeAnnotation]:
        yield from self.package.client.iterate_types()
        for paginator in self.package.paginators:
            yield from paginator.iterate_types()
        for waiter in self.package.waiters:
            yield from waiter.iterate_types()
        if self.package.service_resource:
            yield from self.package.service_resource.iterate_types()
            for collection in self.package.service_resource.collections:
                yield from collection.iterate_types()
            for sub_resource in self.package.service_resource.sub_resources:
                yield from sub_resource.iterate_types()
                for collection in sub_resource.collections:
                    yield from collection.iterate_types()

    def _iterate_types(self) -> Iterator[FakeAnnotation]:
        for type_annotation in self._iterate_types_shallow():
            if isinstance(type_annotation, TypeDefSortable):
                yield from type_annotation.get_children_types()
            else:
                yield type_annotation

    def _replace_external_imports(self) -> None:
        for type_annotation in self._iterate_types():
            if not isinstance(type_annotation, ExternalImport):
                continue

            if type_annotation in self.EXTERNAL_IMPORTS_MAP:
                new_type_annotation = self.EXTERNAL_IMPORTS_MAP[type_annotation]
                type_annotation.copy_from(new_type_annotation)
                continue

            if type_annotation.source.startswith(ImportString("boto3")):
                type_annotation.safe = True
                continue



from typing import Any, Dict, Optional

from botocore.utils import DEFAULT_METADATA_SERVICE_TIMEOUT as DEFAULT_METADATA_SERVICE_TIMEOUT
from botocore.utils import METADATA_BASE_URL as METADATA_BASE_URL
from botocore.utils import (
    ContainerMetadataFetcher,
    IMDSFetcher,
    IMDSRegionProvider,
    InstanceMetadataFetcher,
    InstanceMetadataRegionFetcher,
    S3RegionRedirector,
    S3RegionRedirectorv2,
)
from requests.models import Response

logger: Any
RETRYABLE_HTTP_ERRORS: Any

class AioIMDSFetcher(IMDSFetcher):
    def __init__(
        self,
        timeout: int = ...,
        num_attempts: int = ...,
        base_url: str = ...,
        env: Optional[str] = ...,
        user_agent: Optional[str] = ...,
        config: Optional[Any] = ...,
        session: Optional[Any] = ...,
    ) -> None: ...

class AioInstanceMetadataFetcher(AioIMDSFetcher, InstanceMetadataFetcher):
    async def retrieve_iam_role_credentials(self) -> Dict[str, Any]: ...

class AioIMDSRegionProvider(IMDSRegionProvider):
    async def provide(self) -> str: ...

class AioInstanceMetadataRegionFetcher(AioIMDSFetcher, InstanceMetadataRegionFetcher):
    async def retrieve_region(self) -> Optional[str]: ...  # type: ignore [override]

class AioS3RegionRedirectorv2(S3RegionRedirectorv2):
    async def redirect_from_error(
        self,
        request_dict: Dict[str, Any],
        response: Response,
        operation: Any,
        **kwargs: Any,
    ) -> None: ...
    async def get_bucket_region(self, bucket: Any, response: Response) -> Any: ...

class AioS3RegionRedirector(S3RegionRedirector):
    async def redirect_from_error(
        self, request_dict: Dict[str, Any], response: Response, operation: Any, **kwargs: Any
    ) -> Optional[int]: ...
    async def get_bucket_region(self, bucket: str, response: Response) -> str: ...

class AioContainerMetadataFetcher(ContainerMetadataFetcher):
    def __init__(self, session: Optional[Any] = ..., sleep: Any = ...) -> None: ...
    async def retrieve_full_uri(self, full_url: str, headers: Optional[Any] = ...) -> str: ...
    async def retrieve_uri(self, relative_uri: str) -> Any: ...



"""
EC2 service injected methods.
"""


from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_maps.typed_dicts import TagTypeDef

create_tags_method = Method(
    "create_tags",
    [
        Argument("self", None),
        Argument.kwflag(),
        Argument("Resources", TypeSubscript(Type.Sequence, [Type.str])),
        Argument("Tags", TypeSubscript(Type.Sequence, [TagTypeDef])),
        Argument("DryRun", Type.bool, Type.Ellipsis),
    ],
    Type.none,
)

delete_tags_method = Method(
    "delete_tags",
    [
        Argument("self", None),
        Argument.kwflag(),
        Argument("Resources", TypeSubscript(Type.Sequence, [Type.str])),
        Argument("Tags", TypeSubscript(Type.Sequence, [TagTypeDef]), Type.Ellipsis),
        Argument("DryRun", Type.bool, Type.Ellipsis),
    ],
    Type.none,
)

CLIENT_METHODS = [
    create_tags_method.copy(),
    delete_tags_method.copy(),
]
INSTANCE_METHODS = [
    create_tags_method.copy().remove_argument("Resources"),
    delete_tags_method.copy().remove_argument("Resources"),
]
COMMON_METHODS = [
    create_tags_method.copy().remove_argument("Resources"),
]



from typing import AsyncIterator, Generic, Iterator, TypeVar

from botocore.eventstream import EventStream

_T = TypeVar("_T")

class AioEventStream(Generic[_T], EventStream[_T]):
    def __iter__(self) -> Iterator[_T]: ...
    def __aiter__(self) -> AsyncIterator[_T]: ...
    async def __anext__(self) -> _T: ...
    async def get_initial_response(self) -> _T: ...  # type: ignore [override]



"""
Wrapper for name Union type annotations, like `MyUnion = Union[str, int]`.
"""
from collections.abc import Iterable, Iterator
from typing import TypeVar

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.internal_import_record import InternalImportRecord
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_def_sortable import TypeDefSortable
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript

_R = TypeVar("_R", bound="TypeUnion")


class TypeUnion(TypeSubscript, TypeDefSortable):
    """
    Wrapper for name Union type annotations, like `MyUnion = Union[str, int]`.
    """

    def __init__(
        self,
        children: Iterable[FakeAnnotation],
        name: str = "",
        stringify: bool = False,
    ) -> None:
        self.name = name
        self.parent = Type.Union
        self.children: list[FakeAnnotation] = list(children)
        if len(self.children) < 2:
            raise ValueError("Union must have at least 2 children")
        self._stringify = stringify

    def is_stringified(self) -> bool:
        """
        Whether Union usage should be rendered as a string.
        """
        return self._stringify

    def stringify(self) -> None:
        """
        Render Union usage as a string.
        """
        self._stringify = True

    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        if not self.is_named():
            return super().render(parent_name)

        result = self.name

        if self._stringify:
            result = f'"{result}"'

        return result

    def is_named(self) -> bool:
        """
        Check if type annotation is a named type annotation.
        """
        return self.name != ""

    def copy(self: _R) -> _R:
        """
        Create a copy of type annotation wrapper.
        """
        return self.__class__(
            children=list(self.children),
            name=self.name,
            stringify=self._stringify,
        )

    def debug_render(self) -> str:
        """
        Render type annotation for debug purposes.
        """
        return f"{self}: {', '.join([c.render() for c in self.children])}"

    def get_children_types(self) -> set[FakeAnnotation]:
        """
        Extract required type annotations from attributes.
        """
        result: set[FakeAnnotation] = set()
        for child in self.children:
            result.update(child.iterate_types())
        return result

    def get_sortable_children(self) -> list[TypeDefSortable]:
        """
        Extract required TypeDefSortable list from attributes.
        """
        result: list[TypeDefSortable] = []
        children_types = self.get_children_types()
        for type_annotation in children_types:
            if not isinstance(type_annotation, TypeDefSortable):
                continue
            result.append(type_annotation)

        return result

    def get_children_literals(self, processed: Iterable[str] = ()) -> set[TypeLiteral]:
        """
        Extract required TypeLiteral list from attributes.
        """
        result: set[TypeLiteral] = set()
        if self.name in processed:
            return result
        children_types = self.get_children_types()
        for type_annotation in children_types:
            if isinstance(type_annotation, TypeLiteral):
                result.add(type_annotation)
            if isinstance(type_annotation, TypeDefSortable):
                result.update(type_annotation.get_children_literals((self.name, *processed)))
        return result

    def get_definition_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using Union.
        """
        result: set[ImportRecord] = set()
        result.update(self.parent.get_import_records())
        for child in self.children:
            result.update(child.get_import_records())
        return result

    def _get_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using type annotation.
        """
        return {InternalImportRecord(ServiceModuleName.type_defs, name=self.name)}

    def get_import_records(self) -> set[ImportRecord]:
        """
        Get all import records required for using type annotation.
        """
        return self._get_import_records()

    def iterate_types(self) -> Iterator[FakeAnnotation]:
        """
        Extract type annotations from children.
        """
        if self.is_named():
            yield self
            return

        yield from super().iterate_types()

    def is_type_def(self) -> bool:
        """
        Whether type annotation is a TypeDef.
        """
        return self.is_named()

    @property
    def type_hint_annotations(self) -> list[FakeAnnotation]:
        """
        Type annotations list from arguments and return type with internal types.
        """
        result: list[FakeAnnotation] = []
        for child in self.children:
            if child.get_local_types():
                result.append(child)
        return result

    def is_union(self) -> bool:
        """
        Whether type annotation is a TypeUnion.
        """
        return self.parent == Type.Union

    def iterate_children(self) -> Iterator[FakeAnnotation]:
        """
        Iterate over children.
        """
        yield from self.children



"""
Version manager for PyPI packages.
"""
import json
from json.decoder import JSONDecodeError
from urllib.error import HTTPError
from urllib.request import urlopen

from newversion import Version


class PyPIManager:
    """
    Version manager for PyPI packages.

    Arguments:
        package -- PyPI package name
    """

    JSON_URL = "https://pypi.org/pypi/{package}/json"

    def __init__(self, package: str) -> None:
        self.package = package
        self._versions: set[Version] | None = None

    @property
    def json_url(self) -> str:
        """
        Package JSON URL on PyPI.
        """
        return self.JSON_URL.format(package=self.package)

    def has_version(self, version: str) -> bool:
        """
        Check if version is already uploaded to PyPI.

        Arguments:
            version -- Target version
        """
        return Version(version) in self._get_versions()

    def get_next_version(self, version: str) -> str:
        """
        Get not existing version or closest not existing post-release.

        Arguments:
            version -- Target version
        """
        versions = self._get_versions()
        new_version = Version(version)
        while new_version in versions:
            new_version = new_version.bump_postrelease()
        return new_version.dumps()

    def _get_versions(self) -> set[Version]:
        if self._versions is not None:
            return self._versions

        try:
            with urlopen(self.json_url) as response:
                data_raw = response.read()
        except HTTPError as e:
            if e.code == 404:
                return set()
            raise RuntimeError(f"Cannot retrieve {self.json_url}: {e}") from None

        try:
            data = json.loads(data_raw)
        except JSONDecodeError:
            return set()

        if "releases" not in data:
            return set()

        version_strs = set(data["releases"].keys())
        self._versions = {Version(i) for i in version_strs}
        return self._versions



"""
AioBotocore stubs/docs generator.
"""
from mypy_boto3_builder.constants import TEMPLATES_PATH
from mypy_boto3_builder.generators.base_generator import BaseGenerator
from mypy_boto3_builder.package_data import (
    TypesAioBotocoreLitePackageData,
    TypesAioBotocorePackageData,
)
from mypy_boto3_builder.postprocessors.aiobotocore import AioBotocorePostprocessor
from mypy_boto3_builder.structures.service_package import ServicePackage
from mypy_boto3_builder.utils.version import get_aiobotocore_version
from mypy_boto3_builder.writers.aiobotocore_processors import (
    process_aiobotocore_stubs,
    process_aiobotocore_stubs_docs,
    process_aiobotocore_stubs_lite,
)


class AioBotocoreGenerator(BaseGenerator):
    """
    AioBotocore stubs/docs generator.
    """

    service_package_data = TypesAioBotocorePackageData
    service_template_path = TEMPLATES_PATH / "aiobotocore_service"

    def get_library_version(self) -> str:
        """
        Get underlying library version.
        """
        return get_aiobotocore_version()

    def get_postprocessor(self, service_package: ServicePackage) -> AioBotocorePostprocessor:
        """
        Get postprocessor for service package.
        """
        return AioBotocorePostprocessor(self.session, service_package, self.master_service_names)

    def generate_stubs(self) -> None:
        """
        Generate `aiobotocore-stubs` package.
        """
        self._generate_stubs()
        self._generate_stubs_lite()

    def _generate_stubs(self) -> None:
        package_data = TypesAioBotocorePackageData
        version = self._get_package_version(package_data.PYPI_NAME, self.version)
        if not version:
            return

        self.logger.info(f"Generating {package_data.PYPI_NAME} {version}")
        process_aiobotocore_stubs(
            self.session,
            self.output_path,
            self.master_service_names,
            generate_setup=self.generate_setup,
            version=version,
        )

    def _generate_stubs_lite(self) -> None:
        package_data = TypesAioBotocoreLitePackageData
        version = self._get_package_version(package_data.PYPI_NAME, self.version)
        if not version:
            return

        self.logger.info(f"Generating {package_data.PYPI_NAME} {version}")
        process_aiobotocore_stubs_lite(
            self.session,
            self.output_path,
            self.master_service_names,
            generate_setup=self.generate_setup,
            version=version,
        )

    def generate_docs(self) -> None:
        """
        Generate service and master docs.
        """
        package_data = TypesAioBotocorePackageData
        total_str = f"{len(self.service_names)}"

        self.logger.info(f"Generating {package_data.PYPI_NAME} module docs")
        process_aiobotocore_stubs_docs(
            self.session,
            self.output_path,
            self.service_names,
        )

        for index, service_name in enumerate(self.service_names):
            current_str = f"{{:0{len(total_str)}}}".format(index + 1)
            package_name = package_data.get_service_package_name(service_name)
            self.logger.info(f"[{current_str}/{total_str}] Generating {package_name} module docs")
            self._process_service_docs(
                service_name=service_name,
                package_data=package_data,
                templates_path=TEMPLATES_PATH / "aiobotocore_service_docs",
            )



"""
Wrapper for `typing` type annotation.
"""
from typing import TypeVar

from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation

_R = TypeVar("_R", bound="TypeAnnotation")


class TypeAnnotation(FakeAnnotation):
    """
    Wrapper for `typing` type annotation.

    Arguments:
        wrapped_type -- Original type annotation as a string.
    """

    _typing = ImportString("typing")
    _typing_extensions = ImportString("typing_extensions")

    # Set of supported type annotations. value is default import module
    SUPPORTED_TYPES: dict[str, ImportString] = {
        "Union": _typing,  # typing.Union
        "Any": _typing,  # typing.Any
        "Dict": _typing,  # typing.Dict
        "List": _typing,  # typing.List
        "Set": _typing,  # typing.Set
        "Optional": _typing,  # typing.Optional
        "IO": _typing,  # typing.IO
        "overload": _typing,  # typing.overload
        "Type": _typing,  # typing.Type
        "NoReturn": _typing,  # typing.NoReturn
        "TypedDict": _typing,  # typing_extensions.TypedDict / typing.TypedDict
        "Literal": _typing,  # typing_extensions.Literal / typing.Literal
        "Mapping": _typing,  # typing.Mapping
        "Sequence": _typing,  # typing.Sequence
        "Callable": _typing,  # typing.Callable
        "Iterator": _typing,  # typing.Iterator
        "Awaitable": _typing,  # typing.Awaitable
        "AsyncIterator": _typing,  # typing.AsyncIterator
        "NotRequired": _typing,  # typing_extensions.NotRequired / typing.NotRequired
    }

    # Set of fallback type annotations
    FALLBACK: dict[str, tuple[tuple[int, int] | None, ImportString]] = {
        "NotRequired": ((3, 12), _typing_extensions),
        "TypedDict": ((3, 12), _typing_extensions),
        "Literal": ((3, 12), _typing_extensions),
    }

    def __init__(self, wrapped_type: str) -> None:
        if wrapped_type not in self.SUPPORTED_TYPES:
            raise ValueError(f"Cannot wrap {wrapped_type}")

        self._wrapped_type: str = wrapped_type

    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        return self.get_import_name()

    def get_import_name(self) -> str:
        """
        Create a safe name for imported annotation.
        """
        return self._wrapped_type

    def _get_import_records(self) -> set[ImportRecord]:
        """
        Create a safe Import Record for annotation.
        """
        name = self.get_import_name()
        source = self.SUPPORTED_TYPES[name]
        if name not in self.FALLBACK:
            return {ImportRecord(source=source, name=name)}

        fallback_min_version, fallback_source = self.FALLBACK[name]

        return {
            ImportRecord(
                source=source,
                name=name,
                fallback=ImportRecord(source=fallback_source, name=name),
                min_version=fallback_min_version,
            )
        }

    def is_dict(self) -> bool:
        """
        Whether annotation is a plain Dict.
        """
        return self._wrapped_type == "Dict"

    def is_list(self) -> bool:
        """
        Whether annotation is a plain List.
        """
        return self._wrapped_type == "List"

    def is_union(self) -> bool:
        """
        Whether annotation is a Union.
        """
        return self._wrapped_type == "Union"

    def copy(self: _R) -> _R:
        """
        Create a copy of type annotation wrapper.
        """
        return self.__class__(self._wrapped_type)



"""
Sortable protocol for TypeDefSorter.
"""
from collections.abc import Iterable, Iterator
from typing import Any, Protocol, runtime_checkable

from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral


@runtime_checkable
class TypeDefSortable(Protocol):
    """
    Sortable protocol for TypeDefSorter.
    """

    name: str
    _stringify: bool

    def __gt__(self, other: FakeAnnotation) -> bool:
        ...

    def __lt__(self, other: FakeAnnotation) -> bool:
        ...

    def get_sortable_children(self) -> list["TypeDefSortable"]:
        """
        Extract required sortable TypeDef list from attributes.
        """
        ...

    def is_stringified(self) -> bool:
        """
        Whether TypeDef usage should be rendered as a string.
        """
        ...

    def stringify(self) -> None:
        """
        Render TypeDef usage as a string.
        """
        ...

    def get_children_types(self) -> set[FakeAnnotation]:
        """
        Extract required type annotations from attributes.
        """
        ...

    def debug_render(self) -> str:
        """
        Render type annotation for debug purposes.
        """
        ...

    def get_children_literals(self, processed: Iterable[str] = ()) -> set[TypeLiteral]:
        """
        Extract required TypeLiteral list from attributes.
        """
        ...

    def is_type_def(self) -> bool:
        """
        Whether type annotation is a TypeDef.
        """
        ...

    @property
    def type_hint_annotations(self) -> list[FakeAnnotation]:
        """
        Type annotations list from arguments and return type with internal types.
        """
        ...

    def get_definition_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using TypeAnnotation.
        """
        ...

    def is_typed_dict(self) -> bool:
        """
        Whether type annotation is a TypedDict.
        """
        ...

    def is_union(self) -> bool:
        """
        Whether type annotation is a TypeUnion.
        """
        ...

    def iterate_children(self) -> Iterator[Any]:
        """
        Iterate over children.
        """
        ...



import logging
from typing import Any

from aioboto3.resources.base import AIOBoto3ServiceResource
from boto3.resources.factory import ResourceFactory
from boto3.utils import ServiceContext
from botocore.hooks import BaseEventHooks

logger: logging.Logger

class AIOBoto3ResourceFactory(ResourceFactory):
    def __init__(self, emitter: BaseEventHooks) -> None: ...
    async def load_from_definition(  # type: ignore [override]
        self,
        resource_name: str,
        single_resource_json_definition: Any,
        service_context: ServiceContext,
    ) -> AIOBoto3ServiceResource: ...



"""
Boto3 ServiceResource sub-Resource.
"""
from collections.abc import Iterator

from boto3.resources.base import ServiceResource as Boto3ServiceResource

from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.structures.collection import Collection
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


class Resource(ClassRecord):
    """
    Boto3 ServiceResource sub-Resource.
    """

    def __init__(self, name: str, service_name: ServiceName):
        super().__init__(
            name=name,
            bases=[ExternalImport.from_class(Boto3ServiceResource)],
            use_alias=True,
        )
        self.service_name: ServiceName = service_name
        self.collections: list[Collection] = []

    @property
    def boto3_doc_link(self) -> str:
        """
        Link to boto3 docs.
        """
        return self.service_name.get_boto3_doc_link("ServiceResource", self.name)

    def iterate_types(self) -> Iterator[FakeAnnotation]:
        """
        Iterate over all type annotations from collections.
        """
        yield from super().iterate_types()
        for collection in self.collections:
            yield from collection.iterate_types()



from typing import Any, Dict, List, Union

from boto3.resources.base import ServiceResource
from boto3.resources.response import RawHandler, ResourceHandler

class AIOResourceHandler(ResourceHandler):
    async def __call__(  # type: ignore [override]
        self, parent: ServiceResource, params: Dict[str, Any], response: Dict[str, Any]
    ) -> Union[ServiceResource, List[ServiceResource]]: ...

class AIORawHandler(RawHandler):
    async def __call__(  # type: ignore [override]
        self, parent: ServiceResource, params: Dict[str, Any], response: Dict[str, Any]
    ) -> Dict[str, Any]: ...



"""
S3 service injected methods.
"""

from boto3.s3.transfer import TransferConfig
from botocore.client import BaseClient

from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_maps.named_unions import FileobjTypeDef, PresignedPostConditionsTypeDef
from mypy_boto3_builder.type_maps.typed_dicts import CopySourceTypeDef

callback_arg = Argument(
    "Callback",
    Type.get_optional(TypeSubscript(Type.Callable, [Type.Ellipsis, Type.Any])),
    Type.Ellipsis,
)

config_arg = Argument(
    "Config",
    Type.get_optional(ExternalImport.from_class(TransferConfig)),
    Type.Ellipsis,
)

extra_args_arg = Argument("ExtraArgs", Type.get_optional(Type.DictStrAny), Type.Ellipsis)

source_client_arg = Argument(
    "SourceClient",
    Type.get_optional(ExternalImport.from_class(BaseClient)),
    Type.Ellipsis,
)

copy_method = Method(
    "copy",
    [
        Argument("self", None),
        Argument("CopySource", CopySourceTypeDef),
        Argument("Bucket", Type.str),
        Argument("Key", Type.str),
        extra_args_arg,
        callback_arg,
        source_client_arg,
        config_arg,
    ],
    Type.none,
)

download_file_method = Method(
    "download_file",
    [
        Argument("self", None),
        Argument("Bucket", Type.str),
        Argument("Key", Type.str),
        Argument("Filename", Type.str),
        extra_args_arg,
        callback_arg,
        config_arg,
    ],
    Type.none,
)

download_fileobj_method = Method(
    "download_fileobj",
    [
        Argument("self", None),
        Argument("Bucket", Type.str),
        Argument("Key", Type.str),
        Argument("Fileobj", FileobjTypeDef),
        extra_args_arg,
        callback_arg,
        config_arg,
    ],
    Type.none,
)

generate_presigned_post_method = Method(
    "generate_presigned_post",
    [
        Argument("self", None),
        Argument("Bucket", Type.str),
        Argument("Key", Type.str),
        Argument("Fields", Type.get_optional(Type.DictStrAny), Type.Ellipsis),
        Argument("Conditions", Type.get_optional(PresignedPostConditionsTypeDef), Type.Ellipsis),
        Argument("ExpiresIn", Type.int, TypeConstant(3600)),
    ],
    Type.DictStrAny,
)

upload_file_method = Method(
    "upload_file",
    [
        Argument("self", None),
        Argument("Filename", Type.str),
        Argument("Bucket", Type.str),
        Argument("Key", Type.str),
        extra_args_arg,
        callback_arg,
        config_arg,
    ],
    Type.none,
)

upload_fileobj_method = Method(
    "upload_fileobj",
    [
        Argument("self", None),
        Argument("Fileobj", FileobjTypeDef),
        Argument("Bucket", Type.str),
        Argument("Key", Type.str),
        extra_args_arg,
        callback_arg,
        config_arg,
    ],
    Type.none,
)


CLIENT_METHODS = [
    copy_method,
    download_file_method,
    download_fileobj_method,
    generate_presigned_post_method,
    upload_file_method,
    upload_fileobj_method,
]
BUCKET_METHODS = [
    copy_method.copy().remove_argument("Bucket"),
    download_file_method.copy().remove_argument("Bucket"),
    download_fileobj_method.copy().remove_argument("Bucket"),
    upload_file_method.copy().remove_argument("Bucket"),
    upload_fileobj_method.copy().remove_argument("Bucket"),
]
OBJECT_METHODS = [
    copy_method.copy().remove_argument("Bucket", "Key"),
    download_file_method.copy().remove_argument("Bucket", "Key"),
    download_fileobj_method.copy().remove_argument("Bucket", "Key"),
    upload_file_method.copy().remove_argument("Bucket", "Key"),
    upload_fileobj_method.copy().remove_argument("Bucket", "Key"),
]



"""
DynamoDB service injected methods.
"""

from boto3.dynamodb.table import BatchWriter

from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript

batch_writer_method = Method(
    "batch_writer",
    [
        Argument("self", None),
        Argument(
            "overwrite_by_pkeys",
            TypeSubscript(Type.List, [Type.str]),
            Type.Ellipsis,
        ),
    ],
    ExternalImport.from_class(BatchWriter),
)

TABLE_METHODS = [batch_writer_method]



from logging import Logger
from typing import Any, Mapping, Optional

from botocore.endpoint_provider import RuleSetEndpoint
from botocore.model import OperationModel
from botocore.regions import EndpointRulesetResolver

LOG: Logger

class AioEndpointRulesetResolver(EndpointRulesetResolver):
    async def construct_endpoint(  # type: ignore [override]
        self,
        operation_model: OperationModel,
        call_args: Optional[Mapping[str, Any]],
        request_context: Optional[Mapping[str, Any]],
    ) -> RuleSetEndpoint: ...



"""
Base postprocessor for classes and methods.
"""
from abc import ABC, abstractmethod
from collections.abc import Sequence

from boto3.session import Session

from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.service_package import ServicePackage
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type_typed_dict import TypedDictAttribute, TypeTypedDict
from mypy_boto3_builder.utils.boto3_utils import get_boto3_resource, get_region_name_literal


class BasePostprocessor(ABC):
    """
    Base postprocessor for classes and methods.

    Arguments:
        session -- Boto3 session
        package -- Service package
        service_names -- Available service names
    """

    def __init__(
        self, session: Session, package: ServicePackage, service_names: Sequence[ServiceName]
    ) -> None:
        self.session = session
        self.package = package
        self.service_names = service_names
        self.docs_package_name = self.package.data.PYPI_NAME
        self.logger = get_logger()

    def _has_service_resource(self, service_name: ServiceName) -> bool:
        return bool(get_boto3_resource(self.session, service_name))

    def generate_docstrings(self) -> None:
        """
        Generate all docstrings.
        """
        self._generate_docstrings_client()
        self._generate_docstrings_paginators()
        self._generate_docstrings_waiters()
        self._generate_docstrings_service_resource()
        self._generate_docstrings_collections()
        self._generate_docstrings_sub_resources()

    @abstractmethod
    def process_package(self) -> None:
        """
        Postprocess built package.
        """

    def _generate_docstrings_client(self) -> None:
        boto3_doc_link = self.package.client.boto3_doc_link
        local_doc_link = self.package.get_doc_link("client")
        self.package.client.docstring = (
            f"[Show boto3 documentation]({boto3_doc_link})\n"
            f"[Show {self.docs_package_name} documentation]({local_doc_link})"
        )
        for method in self.package.client.methods:
            docstring = f"{method.docstring}\n\n" if method.docstring else ""
            boto3_doc_link = self.package.service_name.get_boto3_doc_link("Client", method.name)
            local_doc_link = self.package.get_doc_link("client", method.name)
            method.docstring = (
                f"{docstring}"
                f"[Show boto3 documentation]({boto3_doc_link})\n"
                f"[Show {self.docs_package_name} documentation]({local_doc_link})"
            )

    def _generate_docstrings_paginators(self) -> None:
        for paginator in self.package.paginators:
            boto3_doc_link = paginator.boto3_doc_link
            local_doc_link = self.package.get_doc_link("paginators", paginator.name)
            paginator.docstring = (
                f"[Show boto3 documentation]({boto3_doc_link})\n"
                f"[Show {self.docs_package_name} documentation]({local_doc_link})"
            )
            for method in paginator.methods:
                docstring = f"{method.docstring}\n\n" if method.docstring else ""
                boto3_doc_link = self.package.service_name.get_boto3_doc_link(
                    "Paginator", paginator.paginator_name, method.name
                )
                method.docstring = (
                    f"{docstring}"
                    f"[Show boto3 documentation]({boto3_doc_link})\n"
                    f"[Show {self.docs_package_name} documentation]({local_doc_link})"
                )

    def _generate_docstrings_waiters(self) -> None:
        for waiter in self.package.waiters:
            boto3_doc_link = waiter.boto3_doc_link
            local_doc_link = self.package.get_doc_link("waiters", waiter.name)
            waiter.docstring = (
                f"[Show boto3 documentation]({boto3_doc_link})\n"
                f"[Show {self.docs_package_name} documentation]({local_doc_link})"
            )
            for method in waiter.methods:
                docstring = f"{method.docstring}\n\n" if method.docstring else ""
                boto3_doc_link = self.package.service_name.get_boto3_doc_link(
                    "Waiter", waiter.name.replace("Waiter", ""), method.name
                )
                method.docstring = (
                    f"{docstring}"
                    f"[Show boto3 documentation]({boto3_doc_link})\n"
                    f"[Show {self.docs_package_name} documentation]({local_doc_link})"
                )

    def _generate_docstrings_service_resource(self) -> None:
        if not self.package.service_resource:
            return

        boto3_doc_link = self.package.service_resource.boto3_doc_link
        local_doc_link = self.package.get_doc_link("service_resource")
        self.package.service_resource.docstring = (
            f"[Show boto3 documentation]({boto3_doc_link})\n"
            f"[Show {self.docs_package_name} documentation]({local_doc_link})"
        )
        for method in self.package.service_resource.methods:
            docstring = f"{method.docstring}\n\n" if method.docstring else ""
            boto3_doc_link = self.package.service_name.get_boto3_doc_link(
                "ServiceResource", method.name
            )
            local_doc_link = self.package.get_doc_link(
                "service_resource", self.package.service_resource.name, f"{method.name} method"
            )
            method.docstring = (
                f"{docstring}"
                f"[Show boto3 documentation]({boto3_doc_link})\n"
                f"[Show {self.docs_package_name} documentation]({local_doc_link})"
            )

    def _generate_docstrings_collections(self) -> None:
        if not self.package.service_resource:
            return

        for collection in self.package.service_resource.collections:
            boto3_doc_link = collection.boto3_doc_link
            local_doc_link = self.package.get_doc_link("service_resource", collection.name)
            collection.docstring = (
                f"[Show boto3 documentation]({boto3_doc_link})\n"
                f"[Show {self.docs_package_name} documentation]({local_doc_link})"
            )
            for method in collection.methods:
                docstring = f"{method.docstring}\n\n" if method.docstring else ""
                method.docstring = (
                    f"{docstring}"
                    f"[Show boto3 documentation]({boto3_doc_link})\n"
                    f"[Show {self.docs_package_name} documentation]({local_doc_link})"
                )

    def _generate_docstrings_sub_resources(self) -> None:
        if not self.package.service_resource:
            return

        for sub_resource in self.package.service_resource.sub_resources:
            boto3_doc_link = sub_resource.boto3_doc_link
            local_doc_link = self.package.get_doc_link("service_resource", sub_resource.name)
            sub_resource.docstring = (
                f"[Show boto3 documentation]({boto3_doc_link})\n"
                f"[Show {self.docs_package_name} documentation]({local_doc_link})"
            )
            for method in sub_resource.methods:
                docstring = f"{method.docstring}\n\n" if method.docstring else ""
                boto3_doc_link = self.package.service_name.get_boto3_doc_link(
                    sub_resource.name, method.name
                )
                local_doc_link = self.package.get_doc_link(
                    "service_resource", sub_resource.name, f"{method.name} method"
                )
                method.docstring = (
                    f"{docstring}"
                    f"[Show boto3 documentation]({boto3_doc_link})\n"
                    f"[Show {self.docs_package_name} documentation]({local_doc_link})"
                )

    def extend_literals(self) -> None:
        """
        Add extra literals.

        - `<Class>ServiceName`
        - `ServiceName`
        - `ResourceServiceName`
        - `PaginatorName`
        - `WaiterName`
        - `RegionName`
        """
        self.package.literals.append(
            TypeLiteral(
                f"{self.package.service_name.class_name}ServiceName",
                [self.package.service_name.boto3_name],
            )
        )
        self.package.literals.append(
            TypeLiteral("ServiceName", [i.boto3_name for i in self.service_names])
        )
        resource_service_names = [i for i in self.service_names if self._has_service_resource(i)]
        self.package.literals.append(
            TypeLiteral("ResourceServiceName", [i.boto3_name for i in resource_service_names])
        )

        paginator_names = [paginator.operation_name for paginator in self.package.paginators]
        if paginator_names:
            self.package.literals.append(TypeLiteral("PaginatorName", paginator_names))

        waiter_names = [waiter.waiter_name for waiter in self.package.waiters]
        if waiter_names:
            self.package.literals.append(TypeLiteral("WaiterName", waiter_names))

        region_name_literal = get_region_name_literal(self.session, [self.package.service_name])
        if region_name_literal:
            self.package.literals.append(region_name_literal)

    def _replace_typed_dict_with_dict(
        self, attribute: TypedDictAttribute, reference: TypeTypedDict, parent: TypeTypedDict
    ) -> None:
        replacement = Type.DictStrAny
        if attribute.type_annotation is reference:
            attribute.type_annotation = replacement
            self.logger.debug(
                f"Replaced {reference.name} with {replacement.render()} in"
                f" {parent.name}.{attribute.name}"
            )
            return

        if isinstance(attribute.type_annotation, TypeSubscript):
            attribute.type_annotation.replace_child(reference, Type.DictStrAny)
            self.logger.debug(
                f"Deep replaced {reference.name} with {replacement.render()} in"
                f" {parent.name}.{attribute.name}"
            )
            return

        raise RuntimeError(
            f"Cannot replace child {reference.name} in {parent.name}.{attribute.name}"
        )

    def _replace_typed_dict_references(
        self, typed_dict: TypeTypedDict, reference: TypeTypedDict, depth: int
    ) -> None:
        """
        Replace self references with `Dict[str, Any]` to avoid circular dependencies.
        """
        if depth <= 0:
            return
        next_depth = depth - 1

        for child in typed_dict.iterate_children():
            for child_typed_dict in child.type_annotation.iterate_types():
                if not isinstance(child_typed_dict, TypeTypedDict):
                    continue

                if child_typed_dict is reference:
                    self._replace_typed_dict_with_dict(child, reference, typed_dict)
                if next_depth > 0:
                    self._replace_typed_dict_references(child_typed_dict, typed_dict, next_depth)

    def replace_self_ref_typed_dicts(self) -> None:
        """
        Remove self-references from TypedDicts.
        """
        for type_def in self.package.type_defs:
            if isinstance(type_def, TypeTypedDict):
                self._replace_typed_dict_references(type_def, type_def, 2)



from botocore.configprovider import ConfigValueStore, SmartDefaultsConfigStoreFactory

class AioSmartDefaultsConfigStoreFactory(SmartDefaultsConfigStoreFactory):
    async def merge_smart_defaults(self, config_store: ConfigValueStore, mode: str, region_name: str) -> None: ...  # type: ignore [override]
    async def resolve_auto_mode(self, region_name: str) -> str: ...  # type: ignore [override]



"""
Alias map fixes added by botocore for documentation build.

# https://github.com/boto/botocore/blob/develop/botocore/handlers.py#L773
# https://github.com/boto/botocore/blob/develop/botocore/handlers.py#L1055
"""

from mypy_boto3_builder.constants import ALL
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog

ARGUMENT_ALIASES: dict[ServiceName, dict[str, dict[str, str | None]]] = {
    ServiceNameCatalog.cloudsearchdomain: {"Search": {"return": "returnFields"}},
    ServiceNameCatalog.logs: {"CreateExportTask": {"from": "fromTime"}},
    ServiceNameCatalog.ec2: {ALL: {"Filter": "Filters"}},
    ServiceNameCatalog.s3: {
        "PutBucketAcl": {"ContentMD5": None},
        "PutBucketCors": {"ContentMD5": None},
        "PutBucketLifecycle": {"ContentMD5": None},
        "PutBucketLogging": {"ContentMD5": None},
        "PutBucketNotification": {"ContentMD5": None},
        "PutBucketPolicy": {"ContentMD5": None},
        "PutBucketReplication": {"ContentMD5": None},
        "PutBucketRequestPayment": {"ContentMD5": None},
        "PutBucketTagging": {"ContentMD5": None},
        "PutBucketVersioning": {"ContentMD5": None},
        "PutBucketWebsite": {"ContentMD5": None},
        "PutObjectAcl": {"ContentMD5": None},
    },
}


def get_argument_alias(
    service_name: ServiceName, operation_name: str, argument_name: str
) -> str | None:
    """
    Get argument alias for operation.

    Args:
        service_name -- Service name
        operation_name -- Resource operation name
        argument_name -- Argument name

    Returns:
        Argument alias name or None if argument has to be deleted.
    """
    checks = (
        (service_name, operation_name),
        (service_name, ALL),
        (ServiceNameCatalog.all, operation_name),
        (ServiceNameCatalog.all, ALL),
    )
    for current_service_name, current_operation_name in checks:
        if current_service_name not in ARGUMENT_ALIASES:
            continue
        service_type_map = ARGUMENT_ALIASES[current_service_name]

        if current_operation_name not in service_type_map:
            continue
        operation_type_map = service_type_map[current_operation_name]

        if argument_name not in operation_type_map:
            continue

        return operation_type_map[argument_name]

    return argument_name



"""
Helper for Python import strings.
"""
import functools
from typing import TypeVar

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.package_data import Boto3StubsPackageData, TypesAioBotocorePackageData

_R = TypeVar("_R", bound="ImportRecord")


@functools.total_ordering
class ImportRecord:
    """
    Helper for Python import strings.

    Arguments:
        source -- Source of import.
        name -- Import name.
        alias -- Import local name.
        min_version -- Minimum Python version, used for fallback.
        fallback -- Fallback ImportRecord.
    """

    builtins_import_string = ImportString("builtins")
    third_party_import_strings = (
        ImportString("boto3"),
        ImportString("botocore"),
    )

    def __init__(
        self: _R,
        source: ImportString,
        name: str = "",
        alias: str = "",
        min_version: tuple[int, ...] | None = None,
        fallback: _R | None = None,
    ) -> None:
        self.source = source
        self.name = name
        self.alias = alias
        self.min_version = min_version
        self.fallback = fallback

    def __bool__(self) -> bool:
        return not self.is_empty()

    def is_empty(self) -> bool:
        """
        Whether import record is an empty string.
        """
        return not self.source

    @classmethod
    def empty(cls: type[_R]) -> _R:
        """
        Whether import record is an empty string.
        """
        return cls(ImportString.empty())

    def render(self) -> str:
        """
        Get rendered string.
        """
        if self.name and self.alias:
            return f"from {self.source} import {self.name} as {self.alias}"
        if self.name:
            return f"from {self.source} import {self.name}"
        if self.alias:
            return f"import {self.source} as {self.alias}"
        if self.source:
            return f"import {self.source}"

        return ""

    def __str__(self) -> str:
        return self.render()

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ImportRecord):
            return False

        return str(self) == str(other)

    def __gt__(self: _R, other: _R) -> bool:
        if self.fallback is not None and other.fallback is None:
            return True

        if other.fallback is not None and self.fallback is None:
            return False

        if self.source == other.source:
            return self.name > other.name

        if self.is_local() and not other.is_local():
            return True

        if other.is_local() and not self.is_local():
            return False

        if self.is_third_party() and not other.is_third_party():
            return True

        if other.is_third_party() and not self.is_third_party():
            return False

        return self.source > other.source

    def get_local_name(self) -> str:
        """
        Get local import name.
        """
        return self.alias or self.name or self.source.render()

    def is_builtins(self) -> bool:
        """
        Whether import is from Python `builtins` module.
        """
        return self.source.startswith(self.builtins_import_string)

    def is_type_defs(self) -> bool:
        """
        Whether import is from `type_defs` module.
        """
        return self.source.parts[-1] == ServiceModuleName.type_defs.value

    def is_third_party(self) -> bool:
        """
        Whether import is from 3rd party module.
        """
        return any(
            self.source.startswith(third_party_import_string)
            for third_party_import_string in self.third_party_import_strings
        )

    def is_local(self) -> bool:
        """
        Whether import is from local module.
        """
        if not self.source:
            return False

        if self.source.master_name.startswith(Boto3StubsPackageData.SERVICE_PREFIX):
            return True

        if self.source.master_name.startswith(TypesAioBotocorePackageData.SERVICE_PREFIX):
            return True

        if self.is_type_defs():
            return True

        return False

    def needs_sys_fallback(self) -> bool:
        """
        Whether ImportString requires `sys` module.
        """
        return bool(self.fallback and self.min_version)



"""
Resource method map for aio libs.
"""
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript

AIO_RESOURCE_METHOD_MAP: dict[ServiceName, dict[str, dict[str, Method]]] = {
    ServiceNameCatalog.dynamodb: {
        "Table": {
            "batch_writer": Method(
                "batch_writer",
                [
                    Argument("self", None),
                    Argument(
                        "overwrite_by_pkeys",
                        TypeSubscript(Type.Optional, [TypeSubscript(Type.List, [Type.str])]),
                        default=Type.Ellipsis,
                    ),
                    Argument("flush_amount", Type.int, default=Type.Ellipsis),
                    Argument("on_exit_loop_sleep", Type.int, default=Type.Ellipsis),
                ],
                return_type=ExternalImport(
                    ImportString("aioboto3", "dynamodb", "table"), "BatchWriter", safe=True
                ),
            )
        }
    }
}


def get_aio_resource_method(
    service_name: ServiceName, resource_name: str, method_name: str
) -> Method | None:
    """
    Get aio resource method override.

    Arguments:
        service_name -- Service name.
        resource_name -- Parent resource name.
        method_name -- Method name.

    Returns:
        Method structure or None.
    """
    return AIO_RESOURCE_METHOD_MAP.get(service_name, {}).get(resource_name, {}).get(method_name)



import sys
from typing import Any, Dict, Optional, Union

from aiobotocore.endpoint import AioEndpoint
from botocore.args import ClientArgsCreator
from botocore.client import ClientEndpointBridge
from botocore.config import Config
from botocore.credentials import Credentials
from botocore.hooks import BaseEventHooks
from botocore.loaders import Loader
from botocore.model import ServiceModel
from botocore.parsers import ResponseParser
from botocore.serialize import BaseRestSerializer
from botocore.signers import RequestSigner

if sys.version_info >= (3, 12):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

class _GetClientArgsTypeDef(TypedDict):
    serializer: BaseRestSerializer
    endpoint: AioEndpoint
    response_parser: ResponseParser
    event_emitter: BaseEventHooks
    request_signer: RequestSigner
    service_model: ServiceModel
    loader: Loader
    client_config: Config
    partition: Optional[str]
    exceptions_factory: Any

class AioClientArgsCreator(ClientArgsCreator):
    def get_client_args(  # type: ignore [override]
        self,
        service_model: ServiceModel,
        region_name: str,
        is_secure: bool,
        endpoint_url: str,
        verify: Optional[Union[str, bool]],
        credentials: Credentials,
        scoped_config: Optional[Dict[str, Any]],
        client_config: Optional[Config],
        endpoint_bridge: ClientEndpointBridge,
        auth_token: Optional[str] = ...,
        endpoints_ruleset_data: Optional[Dict[str, Any]] = ...,
        partition_data: Optional[Dict[str, Any]] = ...,
    ) -> _GetClientArgsTypeDef: ...



"""
Helpers for parsing methods and attributes.
"""
import inspect
from types import MethodType

from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.type import Type


def get_public_methods(inspect_class: object) -> dict[str, MethodType]:
    """
    Extract public methods from any class.

    Arguments:
        inspect_class -- Inspect class.

    Returns:
        A dictionary of method name and method.
    """
    class_members = inspect.getmembers(inspect_class)
    methods: dict[str, MethodType] = {}
    for name, member in class_members:
        if not inspect.ismethod(member):
            continue

        if name.startswith("_"):
            continue

        methods[name] = member

    return methods


def get_dummy_method(name: str) -> Method:
    """
    Get a dummy method in case we cannot parse it.

    Arguments:
        name -- Method name.

    Returns:
        Method structure.
    """
    return Method(
        name=name,
        arguments=[
            Argument("self", None),
            Argument("args", Type.Any, prefix="*"),
            Argument("kwargs", Type.Any, prefix="**"),
        ],
        return_type=Type.Any,
    )



"""
Wrapper for type annotations imported from 3rd party libraries, like `boto3.service.Service`.
"""
import inspect
from typing import TypeVar

from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation

_R = TypeVar("_R", bound="ExternalImport")


class ExternalImport(FakeAnnotation):
    """
    Wrapper for type annotations imported from 3rd party libraries, like `boto3.service.Service`.

    Arguments:
        source -- Module import string.
        name -- Import name.
        alias -- Import local name.
        safe -- Whether import is wrapped in try-except.
    """

    def __init__(
        self,
        source: ImportString,
        name: str = "",
        alias: str = "",
        safe: bool = False,
    ) -> None:
        self.source: ImportString = source
        self.name: str = name
        self.alias: str = alias
        self.safe: bool = safe

    @classmethod
    def from_class(cls: type[_R], obj: type, alias: str = "", safe: bool = False) -> _R:
        """
        Create an instance from an imported class.

        Arguments:
            value -- Any Class.
            alias -- Local name.
            safe -- Whether import is wrapped in try-except.
        """
        module = inspect.getmodule(obj)
        if module is None:
            raise ValueError(f"Unknown module for {obj}")

        module_name = module.__name__
        return cls(
            source=ImportString.from_str(module_name),
            name=obj.__name__,
            alias=alias,
            safe=safe,
        )

    @property
    def import_record(self) -> ImportRecord:
        """
        Get import record required for using type annotation.
        """
        if self.safe:
            return ImportRecord(
                self.source,
                self.name,
                self.alias,
                min_version=None,
                fallback=ImportRecord(ImportString("builtins"), "object", self.name),
            )
        return ImportRecord(source=self.source, name=self.name, alias=self.alias)

    def __hash__(self) -> int:
        return hash(self.import_record.render())

    def render(self, parent_name: str = "") -> str:
        """
        Get string with local name to use.

        Returns:
            Import record local name.
        """
        return self.import_record.get_local_name()

    def _get_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using type annotation.
        """
        return {self.import_record}

    def copy(self: _R) -> _R:
        """
        Create a copy of type annotation wrapper.
        """
        return self.__class__(self.source, self.name, self.alias, self.safe)

    def copy_from(self: _R, other: _R) -> None:
        """
        Copy all fileds from another instance.
        """
        self.source = other.source
        self.name = other.name
        self.safe = other.safe



"""
Class or module attribute.
"""
from collections.abc import Iterator

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant


class Attribute:
    """
    Class or module attribute.

    Attributes:
        name -- Attribute name.
        type -- Attribute type annotation.
        value -- Attribute value.
        type_ignore -- Add type: ignore comment.
        is_reference -- Whether the attribute parsed from references.
        is_identifier -- Whether the attribute parsed from identifiers.
        is_collection -- Whether the attribute parsed from collections.
    """

    def __init__(
        self,
        name: str,
        type_annotation: FakeAnnotation,
        value: TypeConstant | None = None,
        type_ignore: bool = False,
        is_reference: bool = False,
        is_identifier: bool = False,
        is_collection: bool = False,
    ):
        self.name = name
        self.type_annotation: FakeAnnotation = type_annotation
        self.value: TypeConstant | None = value
        self.type_ignore = type_ignore
        self.is_reference = is_reference
        self.is_identifier = is_identifier
        self.is_collection = is_collection

    def iterate_types(self) -> Iterator[FakeAnnotation]:
        """
        Iterate over all type annotations used.

        Yields:
            Type annotation.
        """
        return self.type_annotation.iterate_types()

    def render(self) -> str:
        """
        Render to a string.
        """
        line = f"{self.name}: {self.type_annotation.render()}"
        if self.type_ignore:
            return f"{line}  # type: ignore"
        return line

    def is_autoload_property(self) -> bool:
        """
        Whether the attribute is an autoload property.
        """
        return not self.is_reference and not self.is_identifier and not self.is_collection



"""
Structure for types-aiobotocore module.
"""

from collections.abc import Iterable

from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.package_data import BasePackageData
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.structures.function import Function
from mypy_boto3_builder.structures.package import Package
from mypy_boto3_builder.structures.service_package import ServicePackage


class AioBotocoreStubsPackage(Package):
    """
    Structure for types-aiobotocore module.
    """

    def __init__(
        self,
        data: type[BasePackageData],
        session_class: ClassRecord | None = None,
        service_names: Iterable[ServiceName] = (),
        service_packages: Iterable[ServicePackage] = (),
        init_functions: Iterable[Function] = (),
    ):
        super().__init__(data)
        self.session_class = session_class or ClassRecord("Session")
        self.service_names = list(service_names)
        self.service_packages = list(service_packages)
        self.init_functions = list(init_functions)

    @property
    def essential_service_names(self) -> list[ServiceName]:
        """
        Service names marked as essential.
        """
        result: list[ServiceName] = []
        for service_name in self.service_names:
            if service_name.is_essential():
                result.append(service_name)
        return result

    def get_session_required_import_records(self) -> list[ImportRecord]:
        """
        Get import records for `session.py[i]`.
        """
        import_records = self.session_class.get_required_import_records()
        return sorted(import_records)

    def get_all_names(self) -> list[str]:
        """
        Get names for `__all__` directive.
        """
        result = [
            "get_session",
            "Session",
        ]
        return sorted(result)



from typing import Any, Dict, Optional, Type

from aiobotocore.httpsession import AIOHTTPSession as AIOHTTPSession
from aiobotocore.response import StreamingBody as StreamingBody
from botocore.endpoint import DEFAULT_TIMEOUT as DEFAULT_TIMEOUT
from botocore.endpoint import MAX_POOL_CONNECTIONS as MAX_POOL_CONNECTIONS
from botocore.endpoint import Endpoint, EndpointCreator
from botocore.model import OperationModel, ServiceModel
from requests.models import Request, Response

async def convert_to_response_dict(
    http_response: Response, operation_model: OperationModel
) -> Dict[str, Any]: ...

class AioEndpoint(Endpoint):
    async def close(self) -> None: ...  # type: ignore [override]
    async def create_request(
        self, params: Any, operation_model: Optional[Any] = ...
    ) -> Request: ...

class AioEndpointCreator(EndpointCreator):
    def create_endpoint(  # type: ignore [override]
        self,
        service_model: ServiceModel,
        region_name: str,
        endpoint_url: str,
        verify: Optional[Any] = ...,
        response_parser_factory: Optional[Any] = ...,
        timeout: int = ...,
        max_pool_connections: int = ...,
        http_session_cls: Type[AIOHTTPSession] = ...,
        proxies: Optional[Any] = ...,
        socket_options: Optional[Any] = ...,
        client_cert: Optional[Any] = ...,
        proxies_config: Optional[Any] = ...,
        connector_args: Optional[Any] = ...,
    ) -> AioEndpoint: ...



import logging
from typing import IO, Any, BinaryIO, Callable, Dict, List, Optional, Sequence

from boto3.s3.transfer import TransferConfig as S3TransferConfig

logger: logging.Logger

def inject_s3_transfer_methods(class_attributes: Sequence[Any], **kwargs: Any) -> None: ...
def inject_object_summary_methods(class_attributes: Sequence[Any], **kwargs: Any) -> None: ...
def inject_bucket_methods(class_attributes: Sequence[Any], **kwargs: Any) -> None: ...
async def object_summary_load(self: Any, *args: Any, **kwargs: Any) -> None: ...
async def download_file(
    self: Any,
    Bucket: str,
    Key: str,
    Filename: str,
    ExtraArgs: Optional[List[str]] = ...,
    Callback: Optional[Callable[[int], None]] = ...,
    Config: Optional[S3TransferConfig] = ...,
) -> None: ...
async def download_fileobj(
    self: Any,
    Bucket: str,
    Key: str,
    Fileobj: IO[Any],
    ExtraArgs: Optional[Dict[str, Any]] = ...,
    Callback: Optional[Callable[[int], None]] = ...,
    Config: Optional[S3TransferConfig] = ...,
) -> None: ...
async def upload_fileobj(
    self: Any,
    Fileobj: BinaryIO,
    Bucket: str,
    Key: str,
    ExtraArgs: Optional[Dict[str, Any]] = ...,
    Callback: Optional[Callable[[int], None]] = ...,
    Config: Optional[S3TransferConfig] = ...,
    Processing: Optional[Callable[[bytes], bytes]] = ...,
) -> Any: ...
async def upload_file(
    self: Any,
    Filename: str,
    Bucket: str,
    Key: str,
    ExtraArgs: Optional[Dict[str, Any]] = ...,
    Callback: Optional[Callable[[int], None]] = ...,
    Config: Optional[S3TransferConfig] = ...,
) -> None: ...
async def copy(
    self: Any,
    CopySource: Any,
    Bucket: str,
    Key: str,
    ExtraArgs: Optional[Dict[str, Any]] = ...,
    Callback: Optional[Callable[[int], None]] = ...,
    SourceClient: Optional[Any] = ...,
    Config: Optional[S3TransferConfig] = ...,
) -> None: ...
async def bucket_load(self: Any, *args: Any, **kwargs: Any) -> None: ...



import logging
from typing import Any, Dict, List, Optional, Type, TypeVar

from boto3.dynamodb.table import TableResource
from botocore.client import BaseClient

logger: logging.Logger

_R = TypeVar("_R")

def register_table_methods(base_classes: List[Any], **kwargs: Any) -> None: ...

class CustomTableResource(TableResource):
    # FIXME: Signature of "batch_writer" incompatible with supertype "TableResource"
    def batch_writer(  # type: ignore [override]
        self,
        overwrite_by_pkeys: Optional[List[str]] = ...,
        flush_amount: int = ...,
        on_exit_loop_sleep: int = ...,
    ) -> "BatchWriter": ...

class BatchWriter:
    def __init__(
        self,
        table_name: str,
        client: BaseClient,
        flush_amount: int = ...,
        overwrite_by_pkeys: Optional[List[str]] = ...,
        on_exit_loop_sleep: int = ...,
    ) -> None: ...
    async def put_item(self, Item: Dict[str, Any]) -> None: ...
    async def delete_item(self, Key: Dict[str, Any]) -> None: ...
    async def __aenter__(self: _R) -> _R: ...
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        tb: Any,
    ) -> None: ...



"""
Collection of TypedDicts added by boto3.
"""
from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type_typed_dict import TypedDictAttribute, TypeTypedDict
from mypy_boto3_builder.type_annotations.type_union import TypeUnion

WaiterConfigTypeDef = TypeTypedDict(
    "WaiterConfigTypeDef",
    [
        TypedDictAttribute("Delay", Type.int, False),
        TypedDictAttribute("MaxAttempts", Type.int, False),
    ],
)

PaginatorConfigTypeDef = TypeTypedDict(
    "PaginatorConfigTypeDef",
    [
        TypedDictAttribute("MaxItems", Type.int, False),
        TypedDictAttribute("PageSize", Type.int, False),
        TypedDictAttribute("StartingToken", Type.str, False),
    ],
)

ResponseMetadataTypeDef = TypeTypedDict(
    "ResponseMetadataTypeDef",
    [
        TypedDictAttribute("RequestId", Type.str, True),
        TypedDictAttribute("HostId", Type.str, True),
        TypedDictAttribute("HTTPStatusCode", Type.int, True),
        TypedDictAttribute("HTTPHeaders", Type.DictStrStr, True),
        TypedDictAttribute("RetryAttempts", Type.int, True),
    ],
)

# s3
CopySourceTypeDef = TypeTypedDict(
    "CopySourceTypeDef",
    [
        TypedDictAttribute("Bucket", Type.str, True),
        TypedDictAttribute("Key", Type.str, True),
        TypedDictAttribute("VersionId", Type.str, False),
    ],
)

# ec2
TagTypeDef = TypeTypedDict(
    "TagTypeDef",
    [
        TypedDictAttribute("Key", Type.str, False),
        TypedDictAttribute("Value", Type.str, False),
    ],
)


# dynamodb
EmptyResponseMetadataTypeDef = TypeTypedDict(
    "EmptyResponseMetadataTypeDef",
    [
        TypedDictAttribute("ResponseMetadata", ResponseMetadataTypeDef, True),
    ],
)

# dynamodb
AttributeValueTypeDef: TypeTypedDict = TypeTypedDict(
    "AttributeValueTypeDef",
    [
        TypedDictAttribute("S", Type.str, False),
        TypedDictAttribute("N", Type.str, False),
        TypedDictAttribute("B", Type.bytes, False),
        TypedDictAttribute("SS", TypeSubscript(Type.Sequence, [Type.str]), False),
        TypedDictAttribute("NS", TypeSubscript(Type.Sequence, [Type.str]), False),
        TypedDictAttribute("BS", TypeSubscript(Type.Sequence, [Type.bytes]), False),
        TypedDictAttribute("M", Type.MappingStrAny, False),
        TypedDictAttribute("L", Type.SequenceAny, False),
        TypedDictAttribute("NULL", Type.bool, False),
        TypedDictAttribute("BOOL", Type.bool, False),
    ],
)

# cloudformation
GetTemplateOutputTypeDef = TypeTypedDict(
    "GetTemplateOutputTypeDef",
    [
        TypedDictAttribute("TemplateBody", Type.DictStrAny, True),
        TypedDictAttribute(
            "StagesAvailable",
            TypeSubscript(
                Type.List,
                [InternalImport(name="TemplateStageType", module_name=ServiceModuleName.literals)],
            ),
            True,
        ),
        TypedDictAttribute("ResponseMetadata", ResponseMetadataTypeDef, True),
    ],
)

# iam
# FIXME: https://github.com/boto/botocore/issues/2992
PolicyDocumentStatementTypeDef = TypeTypedDict(
    "PolicyDocumentStatementTypeDef",
    [
        TypedDictAttribute("Effect", Type.str, True),
        TypedDictAttribute(
            "Resource", TypeUnion((Type.str, TypeSubscript(Type.List, [Type.str]))), True
        ),
        TypedDictAttribute("Sid", Type.str, True),
        TypedDictAttribute(
            "Action", TypeUnion((Type.str, TypeSubscript(Type.List, [Type.str]))), True
        ),
    ],
)

PolicyDocumentDictTypeDef = TypeTypedDict(
    "PolicyDocumentDictTypeDef",
    [
        TypedDictAttribute("Version", Type.str, True),
        TypedDictAttribute(
            "Statement", TypeSubscript(Type.List, [PolicyDocumentStatementTypeDef]), True
        ),
    ],
)



"""
Wrapper for subscript type annotations, like `List[str]`.
"""
from collections.abc import Iterable, Iterator
from typing import TypeVar

from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation

_R = TypeVar("_R", bound="TypeSubscript")


class TypeSubscript(FakeAnnotation):
    """
    Wrapper for subscript type annotations, like `List[str]`.

    Arguments:
        parent -- Parent type annotation.
        children -- Children type annotations.
        stringify -- Convert type annotation to string.
    """

    def __init__(
        self,
        parent: FakeAnnotation,
        children: Iterable[FakeAnnotation] = (),
        stringify: bool = False,
    ) -> None:
        self.parent: FakeAnnotation = parent
        self.children: list[FakeAnnotation] = list(children)
        self._stringify = stringify

    def __hash__(self) -> int:
        return hash(f"{self.parent}.{self.children}")

    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        result = self.parent.render()
        if self.children:
            children = ", ".join([i.render(parent_name) for i in self.children])
            result = f"{result}[{children}]"

        if self._stringify:
            result = f'"{result}"'
        return result

    def get_import_records(self) -> set[ImportRecord]:
        """
        Get all import records required for using type annotation.
        """
        result: set[ImportRecord] = set()
        result.update(self.parent.get_import_records())
        for child in self.children:
            result.update(child.get_import_records())
        return result

    def iterate_types(self) -> Iterator[FakeAnnotation]:
        """
        Extract type annotations from children.
        """
        yield from self.parent.iterate_types()
        for child in self.children:
            yield from child.iterate_types()

    def add_child(self, child: FakeAnnotation) -> None:
        """
        Add new child to Substcript.
        """
        self.children.append(child)

    def is_dict(self) -> bool:
        """
        Whether subscript parent is Dict.
        """
        return self.parent.is_dict()

    def is_list(self) -> bool:
        """
        Whether subscript parent is List.
        """
        return self.parent.is_list()

    def copy(self: _R) -> _R:
        """
        Create a copy of type annotation wrapper.
        """
        return self.__class__(
            parent=self.parent,
            children=list(self.children),
            stringify=self._stringify,
        )

    def get_local_types(self) -> list[FakeAnnotation]:
        """
        Get internal types generated by builder.
        """
        result: list[FakeAnnotation] = []
        for child in self.children:
            result.extend(child.get_local_types())
        return result

    def find_type_annotation_parent(self: _R, type_annotation: FakeAnnotation) -> _R | None:
        """
        Check recursively if child is present in subscript.
        """
        if type_annotation in self.children:
            return self

        type_subscript_children = [i for i in self.children if isinstance(i, self.__class__)]
        for child in type_subscript_children:
            result = child.find_type_annotation_parent(type_annotation)
            if result:
                return result

        return None

    def replace_child(self: _R, child: FakeAnnotation, new_child: FakeAnnotation) -> _R:
        """
        Replace child type annotation with a new one.
        """
        if child not in self.children:
            raise ValueError(f"Child not found: {child}")

        index = self.children.index(child)
        self.children[index] = new_child
        return self



"""
Jinja2 `Environment` manager.
"""
import datetime
from collections.abc import Callable
from typing import Any

import jinja2

from mypy_boto3_builder.constants import BUILDER_REPO_URL, TEMPLATES_PATH
from mypy_boto3_builder.utils.strings import get_anchor_link
from mypy_boto3_builder.utils.version import get_builder_version

__all__ = ["JinjaManager"]


class JinjaManager:
    """
    Jinja2 `Environment` manager.
    """

    _environment = jinja2.Environment(
        loader=jinja2.FileSystemLoader(TEMPLATES_PATH.as_posix()),
        undefined=jinja2.StrictUndefined,
    )

    def __init__(self) -> None:
        self._environment.filters["escape_md"] = self.escape_md  # type: ignore
        self.update_globals(
            builder_version=get_builder_version(),
            current_year=str(datetime.datetime.utcnow().year),
            get_anchor_link=get_anchor_link,
            hasattr=hasattr,
            len=len,
            sorted=sorted,
            repr=repr,
            builder_repo_url=BUILDER_REPO_URL,
        )

    @classmethod
    def update_globals(cls, **kwargs: str | bool | Callable[..., Any]) -> None:
        """
        Update global variables in `jinja2.Environment`.

        Arguments:
            kwargs -- Globals to set.
        """
        cls._environment.globals.update(kwargs)  # type: ignore

    @staticmethod
    def escape_md(value: str) -> str:
        """
        Escape underscore characters.
        """
        return value.replace("_", r"\_")

    def get_environment(self) -> jinja2.Environment:
        """
        Get `jinja2.Environment`.
        """
        return self._environment



from types import TracebackType
from typing import Any, Dict, Optional, Type, TypeVar

from botocore.endpoint import MAX_POOL_CONNECTIONS as MAX_POOL_CONNECTIONS
from requests.models import Request, Response

_R = TypeVar("_R")

class AIOHTTPSession:
    def __init__(
        self,
        verify: bool = ...,
        proxies: Optional[Dict[str, str]] = ...,
        timeout: Optional[float] = ...,
        max_pool_connections: int = ...,
        socket_options: Optional[Any] = ...,
        client_cert: Optional[Any] = ...,
        proxies_config: Optional[Any] = ...,
        connector_args: Optional[Any] = ...,
    ) -> None: ...
    async def __aenter__(self: _R) -> _R: ...
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None: ...
    async def close(self) -> None: ...
    async def send(self, request: Request) -> Response: ...
